# Copyright (C) 2015-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from dataclasses import dataclass
import datetime
import logging
import random
import re
from typing import Any, Dict, Iterable, Iterator, List, Optional, Tuple

from psycopg2 import sql
from psycopg2.extras import execute_values

from swh.core.db import BaseDb
from swh.core.db.db_utils import execute_values_generator
from swh.core.db.db_utils import jsonize as _jsonize
from swh.core.db.db_utils import stored_procedure
from swh.model.hashutil import DEFAULT_ALGORITHMS
from swh.model.model import SHA1_SIZE, OriginVisit, OriginVisitStatus, Sha1Git
from swh.model.swhids import ExtendedObjectType, ObjectType
from swh.storage.interface import ListOrder

logger = logging.getLogger(__name__)


def jsonize(d):
    return _jsonize(dict(d) if d is not None else None)


class QueryBuilder:
    def __init__(self) -> None:
        self.parts = sql.Composed([])
        self.params: List[Any] = []

    def add_query_part(self, query_part, params: List[Any] = []) -> None:
        # add a query part along with its run time parameters
        self.parts += query_part
        self.params += params

    def add_pagination_clause(
        self,
        pagination_key: List[str],
        cursor: Optional[Any],
        direction: Optional[ListOrder],
        limit: Optional[int],
        separator: str = "AND",
    ) -> None:
        """Create and add a pagination clause to the query

        Args:
            pagination_key: Pagination key to be used. Use list of strings
               to support alias fields
            cursor: Pagination cursor as a query parameter
            direction: Sort order
            limit: Limit as a query parameter
            separator: Separator to be used as the prefix for the clause
        """

        # Can be used for queries that require pagination
        if cursor:
            operation = "<" if direction == ListOrder.DESC else ">"  # ASC by default
            # pagination_key and the direction will be given as identifiers
            # and cursor is a run time parameter
            pagination_part = sql.SQL(
                "{separator} {pagination_key} {operation} {cursor}"
            ).format(
                separator=sql.SQL(separator),
                pagination_key=sql.Identifier(*pagination_key),
                operation=sql.SQL(operation),
                cursor=sql.Placeholder(),
            )
            self.add_query_part(pagination_part, params=[cursor])
        # Always use order by and limit with pagination
        if direction:
            operation = "DESC" if direction == ListOrder.DESC else "ASC"
            order_part = sql.SQL("ORDER BY {pagination_key} {operation}").format(
                pagination_key=sql.Identifier(*pagination_key),
                operation=sql.SQL(operation),
            )
            self.add_query_part(order_part)
        if limit is not None:
            limit_part = sql.SQL("LIMIT {limit}").format(limit=sql.Placeholder())
            self.add_query_part(limit_part, params=[limit])

    def get_query(self, db_cursor, separator: str = " ") -> str:
        # To get the query as a string; can be used for debugging
        return self.parts.join(separator).as_string(db_cursor)

    def execute(self, db_cursor, separator: str = " ") -> None:
        # Compose and execute
        query = self.parts.join(separator)
        db_cursor.execute(query, self.params)


@dataclass
class ObjectReferencesPartition:
    table_name: str
    year: int
    week: int
    start: datetime.datetime
    end: datetime.datetime


class Db(BaseDb):
    """Proxy to the SWH DB, with wrappers around stored procedures"""

    ##########################
    # Utilities
    ##########################

    def _get_random_row_from_table(self, table_name, cols, id_col, cur=None):
        random_sha1 = bytes(random.randint(0, 255) for _ in range(SHA1_SIZE))
        cur = self._cursor(cur)
        query = """
            (SELECT {cols} FROM {table} WHERE {id_col} >= %s
             ORDER BY {id_col} LIMIT 1)
            UNION
            (SELECT {cols} FROM {table} WHERE {id_col} < %s
             ORDER BY {id_col} DESC LIMIT 1)
            LIMIT 1
            """.format(
            cols=", ".join(cols), table=table_name, id_col=id_col
        )
        cur.execute(query, (random_sha1, random_sha1))
        row = cur.fetchone()
        if row:
            return row[0]

    ##########################
    # Insertion
    ##########################

    def mktemp_dir_entry(self, entry_type, cur=None):
        self._cursor(cur).execute(
            "SELECT swh_mktemp_dir_entry(%s)", (("directory_entry_%s" % entry_type),)
        )

    @stored_procedure("swh_mktemp_revision")
    def mktemp_revision(self, cur=None):
        pass

    @stored_procedure("swh_mktemp_release")
    def mktemp_release(self, cur=None):
        pass

    @stored_procedure("swh_mktemp_snapshot_branch")
    def mktemp_snapshot_branch(self, cur=None):
        pass

    @stored_procedure("swh_content_add")
    def content_add_from_temp(self, cur=None):
        pass

    @stored_procedure("swh_directory_add")
    def directory_add_from_temp(self, cur=None):
        pass

    @stored_procedure("swh_skipped_content_add")
    def skipped_content_add_from_temp(self, cur=None):
        pass

    @stored_procedure("swh_revision_add")
    def revision_add_from_temp(self, cur=None):
        pass

    @stored_procedure("swh_extid_add")
    def extid_add_from_temp(self, cur=None):
        pass

    @stored_procedure("swh_release_add")
    def release_add_from_temp(self, cur=None):
        pass

    ##########################
    # 'content' table
    ##########################

    def content_update_from_temp(self, keys_to_update, cur=None):
        cur = self._cursor(cur)
        cur.execute(
            """select swh_content_update(ARRAY[%s] :: text[])""" % keys_to_update
        )

    content_get_metadata_keys = [
        "sha1",
        "sha1_git",
        "sha256",
        "blake2s256",
        "length",
        "status",
    ]

    content_add_keys = content_get_metadata_keys + ["ctime"]

    skipped_content_keys = [
        "sha1",
        "sha1_git",
        "sha256",
        "blake2s256",
        "length",
        "reason",
        "status",
        "origin",
    ]

    def content_get_metadata_from_hashes(
        self, hashes: List[bytes], algo: str, cur=None
    ):
        cur = self._cursor(cur)
        assert algo in DEFAULT_ALGORITHMS
        query = f"""
            select {", ".join(self.content_get_metadata_keys)}
            from (values %s) as t (hash)
            inner join content on (content.{algo}=hash)
        """
        yield from execute_values_generator(
            cur,
            query,
            ((hash_,) for hash_ in hashes),
        )

    def content_get_range(self, start, end, limit=None, cur=None) -> Iterator[Tuple]:
        """Retrieve contents within range [start, end]."""
        cur = self._cursor(cur)
        query = """select %s from content
                   where %%s <= sha1 and sha1 <= %%s
                   order by sha1
                   limit %%s""" % ", ".join(
            self.content_get_metadata_keys
        )
        cur.execute(query, (start, end, limit))
        yield from cur

    content_hash_keys = ["sha1", "sha1_git", "sha256", "blake2s256"]

    def content_missing_from_list(self, contents, cur=None):
        cur = self._cursor(cur)

        keys = ", ".join(self.content_hash_keys)
        equality = " AND ".join(
            ("t.%s = c.%s" % (key, key)) for key in self.content_hash_keys
        )

        yield from execute_values_generator(
            cur,
            """
            SELECT %s
            FROM (VALUES %%s) as t(%s)
            WHERE NOT EXISTS (
                SELECT 1 FROM content c
                WHERE %s
            )
            """
            % (keys, keys, equality),
            (tuple(c[key] for key in self.content_hash_keys) for c in contents),
        )

    def content_missing_per_sha1(self, sha1s, cur=None):
        cur = self._cursor(cur)

        yield from execute_values_generator(
            cur,
            """
        SELECT t.sha1 FROM (VALUES %s) AS t(sha1)
        WHERE NOT EXISTS (
            SELECT 1 FROM content c WHERE c.sha1 = t.sha1
        )""",
            ((sha1,) for sha1 in sha1s),
        )

    def content_missing_per_sha1_git(self, contents, cur=None):
        cur = self._cursor(cur)

        yield from execute_values_generator(
            cur,
            """
        SELECT t.sha1_git FROM (VALUES %s) AS t(sha1_git)
        WHERE NOT EXISTS (
            SELECT 1 FROM content c WHERE c.sha1_git = t.sha1_git
        )""",
            ((sha1,) for sha1 in contents),
        )

    content_find_cols = [
        "sha1",
        "sha1_git",
        "sha256",
        "blake2s256",
        "length",
        "ctime",
        "status",
    ]

    def content_find(
        self,
        sha1: Optional[bytes] = None,
        sha1_git: Optional[bytes] = None,
        sha256: Optional[bytes] = None,
        blake2s256: Optional[bytes] = None,
        cur=None,
    ):
        """Find the content optionally on a combination of the following
        checksums sha1, sha1_git, sha256 or blake2s256.

        Args:
            sha1: sha1 content
            git_sha1: the sha1 computed `a la git` sha1 of the content
            sha256: sha256 content
            blake2s256: blake2s256 content

        Returns:
            The tuple (sha1, sha1_git, sha256, blake2s256) if found or None.

        """
        cur = self._cursor(cur)

        checksum_dict = {
            "sha1": sha1,
            "sha1_git": sha1_git,
            "sha256": sha256,
            "blake2s256": blake2s256,
        }

        query_parts = [f"SELECT {','.join(self.content_find_cols)} FROM content WHERE "]
        query_params = []
        where_parts = []
        # Adds only those keys which have values exist
        for algorithm in checksum_dict:
            if checksum_dict[algorithm] is not None:
                where_parts.append(f"{algorithm} = %s")
                query_params.append(checksum_dict[algorithm])

        query_parts.append(" AND ".join(where_parts))
        query = "\n".join(query_parts)
        cur.execute(query, query_params)
        content = cur.fetchall()
        return content

    def content_get_random(self, cur=None):
        return self._get_random_row_from_table("content", ["sha1_git"], "sha1_git", cur)

    ##########################
    # 'skipped_content' table
    ##########################

    def skipped_content_missing(self, contents, cur=None):
        if not contents:
            return []
        cur = self._cursor(cur)

        query = """SELECT * FROM (VALUES %s) AS t (%s)
                   WHERE not exists
                   (SELECT 1 FROM skipped_content s WHERE
                       s.sha1 is not distinct from t.sha1::sha1 and
                       s.sha1_git is not distinct from t.sha1_git::sha1 and
                       s.sha256 is not distinct from t.sha256::bytea);""" % (
            (", ".join("%s" for _ in contents)),
            ", ".join(self.content_hash_keys),
        )
        cur.execute(
            query,
            [tuple(cont[key] for key in self.content_hash_keys) for cont in contents],
        )

        yield from cur

    skipped_content_find_cols = [
        "sha1",
        "sha1_git",
        "sha256",
        "blake2s256",
        "length",
        "status",
        "reason",
        "ctime",
    ]

    def skipped_content_find(
        self,
        sha1: Optional[bytes] = None,
        sha1_git: Optional[bytes] = None,
        sha256: Optional[bytes] = None,
        blake2s256: Optional[bytes] = None,
        cur=None,
    ) -> List[Tuple[Any]]:
        cur = self._cursor(cur)

        checksum_dict = {
            "sha1": sha1,
            "sha1_git": sha1_git,
            "sha256": sha256,
            "blake2s256": blake2s256,
        }

        # XXX: The origin part is untested because of
        # https://gitlab.softwareheritage.org/swh/devel/swh-storage/-/issues/4693
        query_parts = [
            f"""
            SELECT {','.join(self.skipped_content_find_cols)}, origin.url AS origin
            FROM skipped_content
            LEFT JOIN origin ON origin.id = skipped_content.origin
            WHERE
            """
        ]
        query_params = []
        where_parts = []
        # Adds only those keys which have values exist
        for algorithm in checksum_dict:
            if checksum_dict[algorithm] is not None:
                where_parts.append(f"{algorithm} = %s")
                query_params.append(checksum_dict[algorithm])

        query_parts.append(" AND ".join(where_parts))
        query = "\n".join(query_parts)
        cur.execute(query, query_params)
        skipped_contents = cur.fetchall()
        return skipped_contents

    ##########################
    # 'directory*' tables
    ##########################

    def directory_missing_from_list(self, directories, cur=None):
        cur = self._cursor(cur)
        yield from execute_values_generator(
            cur,
            """
            SELECT id FROM (VALUES %s) as t(id)
            WHERE NOT EXISTS (
                SELECT 1 FROM directory d WHERE d.id = t.id
            )
            """,
            ((id,) for id in directories),
        )

    directory_ls_cols = [
        "dir_id",
        "type",
        "target",
        "name",
        "perms",
        "status",
        "sha1",
        "sha1_git",
        "sha256",
        "length",
    ]

    def directory_walk_one(self, directory, cur=None):
        cur = self._cursor(cur)
        cols = ", ".join(self.directory_ls_cols)
        query = "SELECT %s FROM swh_directory_walk_one(%%s)" % cols
        cur.execute(query, (directory,))
        yield from cur

    def directory_walk(self, directory, cur=None):
        cur = self._cursor(cur)
        cols = ", ".join(self.directory_ls_cols)
        query = "SELECT %s FROM swh_directory_walk(%%s)" % cols
        cur.execute(query, (directory,))
        yield from cur

    def directory_entry_get_by_path(self, directory, paths, cur=None):
        """Retrieve a directory entry by path."""
        cur = self._cursor(cur)

        cols = ", ".join(self.directory_ls_cols)
        query = "SELECT %s FROM swh_find_directory_entry_by_path(%%s, %%s)" % cols
        cur.execute(query, (directory, paths))

        data = cur.fetchone()
        if set(data) == {None}:
            return None
        return data

    directory_get_entries_cols = ["type", "target", "name", "perms"]

    def directory_get_entries(self, directory: Sha1Git, cur=None) -> List[Tuple]:
        cur = self._cursor(cur)
        cur.execute(
            "SELECT * FROM swh_directory_get_entries(%s::sha1_git)", (directory,)
        )
        return list(cur)

    def directory_get_raw_manifest(
        self, directory_ids: List[Sha1Git], cur=None
    ) -> Iterable[Tuple[Sha1Git, bytes]]:
        cur = self._cursor(cur)
        yield from execute_values_generator(
            cur,
            """
            SELECT t.id, raw_manifest FROM (VALUES %s) as t(id)
            INNER JOIN directory ON (t.id=directory.id)
            """,
            ((id_,) for id_ in directory_ids),
        )

    def directory_get_id_range(
        self, start, end, limit=None, cur=None
    ) -> Iterator[Tuple[Sha1Git]]:
        cur = self._cursor(cur)
        cur.execute(
            """
            SELECT id FROM directory
            WHERE %s <= id AND id <= %s
            LIMIT %s
            """,
            (start, end, limit),
        )
        yield from cur

    def directory_get_random(self, cur=None):
        return self._get_random_row_from_table("directory", ["id"], "id", cur)

    ##########################
    # 'revision' table
    ##########################

    def revision_missing_from_list(self, revisions, cur=None):
        cur = self._cursor(cur)

        yield from execute_values_generator(
            cur,
            """
            SELECT id FROM (VALUES %s) as t(id)
            WHERE NOT EXISTS (
                SELECT 1 FROM revision r WHERE r.id = t.id
            )
            """,
            ((id,) for id in revisions),
        )

    revision_add_cols = [
        "id",
        "date",
        "date_offset",
        "date_neg_utc_offset",
        "date_offset_bytes",
        "committer_date",
        "committer_date_offset",
        "committer_date_neg_utc_offset",
        "committer_date_offset_bytes",
        "type",
        "directory",
        "message",
        "author_fullname",
        "author_name",
        "author_email",
        "committer_fullname",
        "committer_name",
        "committer_email",
        "metadata",
        "synthetic",
        "extra_headers",
        "raw_manifest",
    ]

    revision_get_cols = revision_add_cols + ["parents"]

    @staticmethod
    def mangle_query_key(key, main_table, id_col, ignore_displayname=False):
        if key == "id":
            return id_col
        if key == "parents":
            return f"""
            ARRAY(
            SELECT rh.parent_id::bytea
            FROM revision_history rh
            WHERE rh.id = {id_col}
            ORDER BY rh.parent_rank
            )"""

        if "_" not in key:
            return f"{main_table}.{key}"

        head, tail = key.split("_", 1)
        if head not in ("author", "committer") or tail not in (
            "name",
            "email",
            "id",
            "fullname",
        ):
            return f"{main_table}.{key}"

        if ignore_displayname:
            return f"{head}.{tail}"
        else:
            if tail == "id":
                return f"{head}.{tail}"
            elif tail in ("name", "email"):
                # These fields get populated again from fullname by
                # converters.db_to_author if they're None, so we can just NULLify them
                # when displayname is set.
                return (
                    f"CASE"
                    f" WHEN {head}.displayname IS NULL THEN {head}.{tail} "
                    f" ELSE NULL "
                    f"END AS {key}"
                )
            elif tail == "fullname":
                return f"COALESCE({head}.displayname, {head}.fullname) AS {key}"

        assert False, "All cases should have been handled here"

    def revision_get_from_list(self, revisions, ignore_displayname=False, cur=None):
        cur = self._cursor(cur)

        query_keys = ", ".join(
            self.mangle_query_key(k, "revision", "t.id", ignore_displayname)
            for k in self.revision_get_cols
        )

        yield from execute_values_generator(
            cur,
            """
            SELECT %s FROM (VALUES %%s) as t(sortkey, id)
            LEFT JOIN revision ON t.id = revision.id
            LEFT JOIN person author ON revision.author = author.id
            LEFT JOIN person committer ON revision.committer = committer.id
            ORDER BY sortkey
            """
            % query_keys,
            ((sortkey, id) for sortkey, id in enumerate(revisions)),
        )

    def revision_get_range(self, start, end, limit, cur=None) -> Iterator[Tuple]:
        cur = self._cursor(cur)

        query_keys = ", ".join(
            self.mangle_query_key(k, "revision", "revision.id")
            for k in self.revision_get_cols
        )

        cur.execute(
            """
            SELECT %s FROM revision
            LEFT JOIN person author ON revision.author = author.id
            LEFT JOIN person committer ON revision.committer = committer.id
            WHERE %%s <= revision.id AND revision.id <= %%s
            LIMIT %%s
            """
            % query_keys,
            (start, end, limit),
        )
        yield from cur

    def revision_log(
        self, root_revisions, ignore_displayname=False, limit=None, cur=None
    ):
        cur = self._cursor(cur)

        query = """\
        SELECT %s
        FROM swh_revision_log(
          "root_revisions" := %%s, num_revs := %%s, "ignore_displayname" := %%s
        )""" % ", ".join(
            self.revision_get_cols
        )

        cur.execute(query, (root_revisions, limit, ignore_displayname))
        yield from cur

    revision_shortlog_cols = ["id", "parents"]

    def revision_shortlog(self, root_revisions, limit=None, cur=None):
        cur = self._cursor(cur)

        query = """SELECT %s
                   FROM swh_revision_list(%%s, %%s)
                """ % ", ".join(
            self.revision_shortlog_cols
        )

        cur.execute(query, (root_revisions, limit))
        yield from cur

    def revision_get_random(self, cur=None):
        return self._get_random_row_from_table("revision", ["id"], "id", cur)

    ##########################
    # 'extid' table
    ##########################

    extid_cols = ["extid", "extid_version", "extid_type", "target", "target_type"]

    def extid_get_from_extid_list(
        self, extid_type: str, ids: List[bytes], version: Optional[int] = None, cur=None
    ):
        cur = self._cursor(cur)
        query_keys = ", ".join(
            self.mangle_query_key(k, "extid", "t.id") for k in self.extid_cols
        )
        filter_query = ""
        if version is not None:
            filter_query = cur.mogrify(
                f"WHERE extid_version={version}", (version,)
            ).decode()

        sql = f"""
            SELECT {query_keys}
            FROM (VALUES %s) as t(sortkey, extid, extid_type)
            LEFT JOIN extid USING (extid, extid_type)
            {filter_query}
            ORDER BY sortkey
            """

        yield from execute_values_generator(
            cur,
            sql,
            (((sortkey, extid, extid_type) for sortkey, extid in enumerate(ids))),
        )

    def extid_get_from_swhid_list(
        self,
        target_type: str,
        ids: List[bytes],
        extid_version: Optional[int] = None,
        extid_type: Optional[str] = None,
        cur=None,
    ):
        cur = self._cursor(cur)
        target_type = ObjectType(
            target_type
        ).name.lower()  # aka "rev" -> "revision", ...
        query_keys = ", ".join(
            self.mangle_query_key(k, "extid", "t.id") for k in self.extid_cols
        )
        filter_query = ""
        if extid_version is not None and extid_type is not None:
            filter_query = cur.mogrify(
                "WHERE extid_version=%s AND extid_type=%s",
                (
                    extid_version,
                    extid_type,
                ),
            ).decode()

        sql = f"""
            SELECT {query_keys}
            FROM (VALUES %s) as t(sortkey, target, target_type)
            LEFT JOIN extid USING (target, target_type)
            {filter_query}
            ORDER BY sortkey
            """

        yield from execute_values_generator(
            cur,
            sql,
            (((sortkey, target, target_type) for sortkey, target in enumerate(ids))),
            template=b"(%s,%s,%s::object_type)",
        )

    def extid_delete_for_target(
        self, target_rows: List[Tuple[str, bytes]], cur=None
    ) -> Dict[str, int]:
        result = {}
        cur = self._cursor(cur)
        execute_values(
            cur,
            """DELETE FROM extid
                USING (VALUES %s) AS t(target_type, target)
                WHERE extid.target_type = t.target_type
                  AND extid.target = t.target""",
            target_rows,
            template=b"(%s::object_type,%s)",
        )
        result["extid:delete"] = cur.rowcount
        return result

    ##########################
    # 'release' table
    ##########################

    def release_missing_from_list(self, releases, cur=None):
        cur = self._cursor(cur)
        yield from execute_values_generator(
            cur,
            """
            SELECT id FROM (VALUES %s) as t(id)
            WHERE NOT EXISTS (
                SELECT 1 FROM release r WHERE r.id = t.id
            )
            """,
            ((id,) for id in releases),
        )

    release_add_cols = [
        "id",
        "target",
        "target_type",
        "date",
        "date_offset",
        "date_neg_utc_offset",
        "date_offset_bytes",
        "name",
        "comment",
        "synthetic",
        "raw_manifest",
        "author_fullname",
        "author_name",
        "author_email",
    ]
    release_get_cols = release_add_cols

    def release_get_from_list(self, releases, ignore_displayname=False, cur=None):
        cur = self._cursor(cur)
        query_keys = ", ".join(
            self.mangle_query_key(k, "release", "t.id", ignore_displayname)
            for k in self.release_get_cols
        )

        yield from execute_values_generator(
            cur,
            """
            SELECT %s FROM (VALUES %%s) as t(sortkey, id)
            LEFT JOIN release ON t.id = release.id
            LEFT JOIN person author ON release.author = author.id
            ORDER BY sortkey
            """
            % query_keys,
            ((sortkey, id) for sortkey, id in enumerate(releases)),
        )

    def release_get_range(self, start, end, limit, cur=None) -> Iterator[Tuple]:
        cur = self._cursor(cur)

        query_keys = ", ".join(
            self.mangle_query_key(k, "release", "release.id")
            for k in self.release_get_cols
        )

        cur.execute(
            """
            SELECT %s FROM release
            LEFT JOIN person author ON release.author = author.id
            WHERE %%s <= release.id AND release.id <= %%s
            LIMIT %%s
            """
            % query_keys,
            (start, end, limit),
        )
        yield from cur

    def release_get_random(self, cur=None):
        return self._get_random_row_from_table("release", ["id"], "id", cur)

    ##########################
    # 'snapshot' table
    ##########################

    def snapshot_exists(self, snapshot_id, cur=None):
        """Check whether a snapshot with the given id exists"""
        cur = self._cursor(cur)

        cur.execute("""SELECT 1 FROM snapshot where id=%s""", (snapshot_id,))

        return bool(cur.fetchone())

    def snapshot_missing_from_list(self, snapshots, cur=None):
        cur = self._cursor(cur)
        yield from execute_values_generator(
            cur,
            """
            SELECT id FROM (VALUES %s) as t(id)
            WHERE NOT EXISTS (
                SELECT 1 FROM snapshot d WHERE d.id = t.id
            )
                """,
            ((id,) for id in snapshots),
        )

    def snapshot_add(self, snapshot_id, cur=None):
        """Add a snapshot from the temporary table"""
        cur = self._cursor(cur)

        cur.execute("""SELECT swh_snapshot_add(%s)""", (snapshot_id,))

    snapshot_count_cols = ["target_type", "count"]

    def snapshot_count_branches(
        self,
        snapshot_id,
        branch_name_exclude_prefix=None,
        cur=None,
    ):
        cur = self._cursor(cur)
        query = """\
           SELECT %s FROM swh_snapshot_count_branches(%%s, %%s)
        """ % ", ".join(
            self.snapshot_count_cols
        )

        cur.execute(query, (snapshot_id, branch_name_exclude_prefix))

        yield from cur

    snapshot_get_cols = ["snapshot_id", "name", "target", "target_type"]

    def snapshot_get_by_id(
        self,
        snapshot_id,
        branches_from=b"",
        branches_count=None,
        target_types=None,
        branch_name_include_substring=None,
        branch_name_exclude_prefix=None,
        cur=None,
    ):
        cur = self._cursor(cur)
        query = """\
        SELECT %s
        FROM swh_snapshot_get_by_id(%%s, %%s, %%s, %%s :: snapshot_target[], %%s, %%s)
        """ % ", ".join(
            self.snapshot_get_cols
        )

        cur.execute(
            query,
            (
                snapshot_id,
                branches_from,
                branches_count,
                target_types,
                branch_name_include_substring,
                branch_name_exclude_prefix,
            ),
        )

        yield from cur

    def snapshot_branch_get_by_name(
        self,
        cols_to_fetch,
        snapshot_id,
        branch_name,
        cur=None,
    ):
        cur = self._cursor(cur)
        query = f"""SELECT {", ".join(cols_to_fetch)}
            FROM snapshot_branch sb
            LEFT JOIN snapshot_branches sbs ON sb.object_id = sbs.branch_id
            LEFT JOIN snapshot ss ON sbs.snapshot_id = ss.object_id
            WHERE ss.id=%s AND sb.name=%s
        """
        cur.execute(query, (snapshot_id, branch_name))
        return cur.fetchone()

    def snapshot_get_id_range(
        self, start, end, limit=None, cur=None
    ) -> Iterator[Tuple[Sha1Git]]:
        cur = self._cursor(cur)
        cur.execute(
            """
            SELECT id FROM snapshot
            WHERE %s <= id AND id <= %s
            LIMIT %s
            """,
            (start, end, limit),
        )

        yield from cur

    def snapshot_get_random(self, cur=None):
        return self._get_random_row_from_table("snapshot", ["id"], "id", cur)

    ##########################
    # 'origin_visit' and 'origin_visit_status' tables
    ##########################

    def origin_visit_add(self, origin, ts, type, cur=None):
        """Add a new origin_visit for origin origin at timestamp ts.

        Args:
            origin: origin concerned by the visit
            ts: the date of the visit
            type: type of loader for the visit

        Returns:
            The new visit index step for that origin

        """
        cur = self._cursor(cur)
        self._cursor(cur).execute(
            "SELECT swh_origin_visit_add(%s, %s, %s)", (origin, ts, type)
        )
        return cur.fetchone()[0]

    origin_visit_status_cols = [
        "origin",
        "visit",
        "date",
        "type",
        "status",
        "snapshot",
        "metadata",
    ]

    def origin_visit_status_add(
        self, visit_status: OriginVisitStatus, cur=None
    ) -> None:
        """Add new origin visit status"""
        assert self.origin_visit_status_cols[0] == "origin"
        assert self.origin_visit_status_cols[-1] == "metadata"
        cols = self.origin_visit_status_cols[1:-1]
        cur = self._cursor(cur)
        cur.execute(
            f"WITH origin_id as (select id from origin where url=%s) "
            f"INSERT INTO origin_visit_status "
            f"(origin, {', '.join(cols)}, metadata) "
            f"VALUES ((select id from origin_id), "
            f"{', '.join(['%s']*len(cols))}, %s) "
            f"ON CONFLICT (origin, visit, date) do nothing",
            [visit_status.origin]
            + [getattr(visit_status, key) for key in cols]
            + [jsonize(visit_status.metadata)],
        )

    origin_visit_cols = ["origin", "visit", "date", "type"]

    def origin_visit_add_with_id(self, origin_visit: OriginVisit, cur=None) -> None:
        """Insert origin visit when id are already set"""
        ov = origin_visit
        assert ov.visit is not None
        cur = self._cursor(cur)
        query = """INSERT INTO origin_visit ({cols})
                   VALUES ((select id from origin where url=%s), {values})
                   ON CONFLICT (origin, visit) DO NOTHING""".format(
            cols=", ".join(self.origin_visit_cols),
            values=", ".join("%s" for col in self.origin_visit_cols[1:]),
        )
        cur.execute(query, (ov.origin, ov.visit, ov.date, ov.type))

    origin_visit_get_cols = [
        "origin",
        "visit",
        "date",
        "type",
        "status",
        "metadata",
        "snapshot",
    ]
    origin_visit_select_cols = [
        "o.url AS origin",
        "ov.visit",
        "ov.date",
        "ov.type AS type",
        "ovs.status",
        "ovs.snapshot",
        "ovs.metadata",
    ]

    origin_visit_status_select_cols = [
        "o.url AS origin",
        "ovs.visit",
        "ovs.date",
        "ovs.type AS type",
        "ovs.status",
        "ovs.snapshot",
        "ovs.metadata",
    ]

    def _make_origin_visit_status(
        self, row: Optional[Tuple[Any]]
    ) -> Optional[Dict[str, Any]]:
        """Make an origin_visit_status dict out of a row"""
        if not row:
            return None
        return dict(zip(self.origin_visit_status_cols, row))

    def origin_visit_status_get_latest(
        self,
        origin_url: str,
        visit: int,
        allowed_statuses: Optional[List[str]] = None,
        require_snapshot: bool = False,
        cur=None,
    ) -> Optional[Dict[str, Any]]:
        """Given an origin visit id, return its latest origin_visit_status"""
        cur = self._cursor(cur)

        query_parts = [
            "SELECT %s" % ", ".join(self.origin_visit_status_select_cols),
            "FROM origin_visit_status ovs ",
            "INNER JOIN origin o ON o.id = ovs.origin",
        ]
        query_parts.append("WHERE o.url = %s")
        query_params: List[Any] = [origin_url]
        query_parts.append("AND ovs.visit = %s")
        query_params.append(visit)

        if require_snapshot:
            query_parts.append("AND ovs.snapshot is not null")

        if allowed_statuses:
            query_parts.append("AND ovs.status IN %s")
            query_params.append(tuple(allowed_statuses))

        query_parts.append("ORDER BY ovs.date DESC LIMIT 1")
        query = "\n".join(query_parts)

        cur.execute(query, tuple(query_params))
        row = cur.fetchone()
        return self._make_origin_visit_status(row)

    def origin_visit_status_get_range(
        self,
        origin: str,
        visit: int,
        date_from: Optional[datetime.datetime],
        order: ListOrder,
        limit: int,
        cur=None,
    ):
        """Retrieve visit_status rows for visit (origin, visit) in a paginated way."""
        cur = self._cursor(cur)

        query_parts = [
            f"SELECT {', '.join(self.origin_visit_status_select_cols)} "
            "FROM origin_visit_status ovs ",
            "INNER JOIN origin o ON o.id = ovs.origin ",
        ]
        query_parts.append("WHERE o.url = %s AND ovs.visit = %s ")
        query_params: List[Any] = [origin, visit]

        if date_from is not None:
            op_comparison = ">=" if order == ListOrder.ASC else "<="
            query_parts.append(f"and ovs.date {op_comparison} %s ")
            query_params.append(date_from)

        if order == ListOrder.ASC:
            query_parts.append("ORDER BY ovs.date ASC ")
        elif order == ListOrder.DESC:
            query_parts.append("ORDER BY ovs.date DESC ")
        else:
            assert False

        query_parts.append("LIMIT %s")
        query_params.append(limit)

        query = "\n".join(query_parts)
        cur.execute(query, tuple(query_params))
        yield from cur

    def origin_visit_get_range(
        self,
        origin: str,
        visit_from: int,
        order: ListOrder,
        limit: int,
        cur=None,
    ):
        origin_visit_cols = ["o.url as origin", "ov.visit", "ov.date", "ov.type"]
        builder = QueryBuilder()
        builder.add_query_part(
            sql.SQL(
                f"""SELECT { ', '.join(origin_visit_cols) } FROM origin_visit ov
                INNER JOIN origin o ON o.id = ov.origin
                WHERE ov.origin = (select id from origin where url = %s)"""
            ),
            params=[origin],  # dynamic params
        )
        builder.add_pagination_clause(
            pagination_key=["ov", "visit"],
            cursor=visit_from,
            direction=order,
            limit=limit,
        )
        cur = self._cursor(cur)
        builder.execute(db_cursor=cur)
        yield from cur

    def origin_visit_status_get_all_in_range(
        self,
        origin: str,
        allowed_statuses: Optional[List[str]],
        require_snapshot: bool,
        visit_from: int,
        visit_to: int,
        cur=None,
    ):
        cur = self._cursor(cur)

        query_parts = [
            f"SELECT {', '.join(self.origin_visit_status_select_cols)}",
            " FROM origin_visit_status ovs",
            " INNER JOIN origin o ON o.id = ovs.origin",
        ]
        query_parts.append("WHERE ovs.origin = (select id from origin where url = %s)")
        query_params: List[Any] = [origin]

        assert visit_from <= visit_to

        query_parts.append("AND ovs.visit >= %s")
        query_params.append(visit_from)

        query_parts.append("AND ovs.visit <= %s")
        query_params.append(visit_to)

        if require_snapshot:
            query_parts.append("AND ovs.snapshot is not null")

        if allowed_statuses:
            query_parts.append("AND ovs.status IN %s")
            query_params.append(tuple(allowed_statuses))

        query_parts.append("ORDER BY ovs.visit ASC, ovs.date ASC")

        query = "\n".join(query_parts)
        cur.execute(query, tuple(query_params))
        yield from cur

    def origin_visit_get(self, origin_id, visit_id, cur=None):
        """Retrieve information on visit visit_id of origin origin_id.

        Args:
            origin_id: the origin concerned
            visit_id: The visit step for that origin

        Returns:
            The origin_visit information

        """
        cur = self._cursor(cur)

        query = """\
            SELECT %s
            FROM origin_visit ov
            INNER JOIN origin o ON o.id = ov.origin
            INNER JOIN origin_visit_status ovs USING (origin, visit)
            WHERE ov.origin = (select id from origin where url = %%s) AND ov.visit = %%s
            ORDER BY ovs.date DESC
            LIMIT 1
            """ % (
            ", ".join(self.origin_visit_select_cols)
        )

        cur.execute(query, (origin_id, visit_id))
        r = cur.fetchall()
        if not r:
            return None
        return r[0]

    def origin_visit_find_by_date(self, origin, visit_date, type=None, cur=None):
        cur = self._cursor(cur)
        cur.execute(
            "SELECT * FROM swh_visit_find_by_date(%s, %s, %s)",
            (origin, visit_date, type),
        )
        rows = cur.fetchall()
        if rows:
            visit = dict(zip(self.origin_visit_get_cols, rows[0]))
            visit["origin"] = origin
            return visit

    def origin_visit_exists(self, origin_id, visit_id, cur=None):
        """Check whether an origin visit with the given ids exists"""
        cur = self._cursor(cur)

        query = "SELECT 1 FROM origin_visit where origin = %s AND visit = %s"

        cur.execute(query, (origin_id, visit_id))

        return bool(cur.fetchone())

    def origin_visit_get_latest(
        self,
        origin_id: str,
        type: Optional[str],
        allowed_statuses: Optional[Iterable[str]],
        require_snapshot: bool,
        cur=None,
    ):
        """Retrieve the most recent origin_visit of the given origin,
        with optional filters.

        Args:
            origin_id: the origin concerned
            type: Optional visit type to filter on
            allowed_statuses: the visit statuses allowed for the returned visit
            require_snapshot (bool): If True, only a visit with a known
                snapshot will be returned.

        Returns:
            The origin_visit information, or None if no visit matches.
        """
        cur = self._cursor(cur)

        # using 'LEFT JOIN origin_visit' instead of an INNER JOIN as a query
        # optimization; end results are equivalent.
        query_parts = [
            "SELECT %s" % ", ".join(self.origin_visit_select_cols),
            "FROM origin_visit_status ovs ",
            "INNER JOIN origin o ON o.id = ovs.origin",
            "LEFT JOIN origin_visit ov USING (origin, visit)",
        ]
        query_parts.append(
            "WHERE ovs.origin = (SELECT id FROM origin o WHERE o.url = %s)"
        )
        query_params: List[Any] = [origin_id]

        if type is not None:
            query_parts.append("AND ovs.type = %s")
            query_params.append(type)

        if require_snapshot:
            query_parts.append("AND ovs.snapshot is not null")

        if allowed_statuses:
            query_parts.append("AND ovs.status IN %s")
            query_params.append(tuple(allowed_statuses))

        query_parts.append("ORDER BY ovs.visit DESC, ovs.date DESC LIMIT 1")

        query = "\n".join(query_parts)

        cur.execute(query, tuple(query_params))
        r = cur.fetchone()
        if not r:
            return None
        return r

    def origin_visit_get_random(self, type, cur=None):
        """Randomly select one origin visit that was full and in the last 3
        months

        """
        cur = self._cursor(cur)
        columns = ",".join(self.origin_visit_select_cols)
        query = f"""select {columns}
                    from origin_visit ov
                    inner join origin o on ov.origin=o.id
                    inner join origin_visit_status ovs using (origin, visit)
                    where ovs.status='full'
                      and ov.type=%s
                      and ov.date > now() - '3 months'::interval
                      and random() < 0.1
                    limit 1
                 """
        cur.execute(query, (type,))
        return cur.fetchone()

    ##########################
    # 'origin' table
    ##########################

    def origin_add(self, url, cur=None):
        """Insert a new origin and return the new identifier."""
        insert = """INSERT INTO origin (url) values (%s)
                    ON CONFLICT DO NOTHING
                    """

        cur.execute(insert, (url,))
        return cur.rowcount

    origin_cols = ["url"]

    def origin_get_by_url(self, origins, cur=None):
        """Retrieve origin `(type, url)` from urls if found."""
        cur = self._cursor(cur)

        query = """SELECT %s FROM (VALUES %%s) as t(url)
                   LEFT JOIN origin ON t.url = origin.url
                """ % ",".join(
            "origin." + col for col in self.origin_cols
        )

        yield from execute_values_generator(cur, query, ((url,) for url in origins))

    def origin_get_by_sha1(self, sha1s, cur=None):
        """Retrieve origin urls from sha1s if found."""
        cur = self._cursor(cur)

        query = """SELECT %s FROM (VALUES %%s) as t(sha1)
                   LEFT JOIN origin ON t.sha1 = digest(origin.url, 'sha1')
                """ % ",".join(
            "origin." + col for col in self.origin_cols
        )

        yield from execute_values_generator(cur, query, ((sha1,) for sha1 in sha1s))

    def origin_id_get_by_url(self, origins, cur=None):
        """Retrieve origin `(type, url)` from urls if found."""
        cur = self._cursor(cur)

        query = """SELECT id FROM (VALUES %s) as t(url)
                   LEFT JOIN origin ON t.url = origin.url
                """

        for row in execute_values_generator(cur, query, ((url,) for url in origins)):
            yield row[0]

    origin_get_range_cols = ["id", "url"]

    def origin_get_range(self, origin_from: int = 1, origin_count: int = 100, cur=None):
        """Retrieve ``origin_count`` origins whose ids are greater
        or equal than ``origin_from``.

        Origins are sorted by id before retrieving them.

        Args:
            origin_from: the minimum id of origins to retrieve
            origin_count: the maximum number of origins to retrieve

        """
        cur = self._cursor(cur)

        query = """SELECT %s
                   FROM origin WHERE id >= %%s
                   ORDER BY id LIMIT %%s
                """ % ",".join(
            self.origin_get_range_cols
        )

        cur.execute(query, (origin_from, origin_count))
        yield from cur

    def _origin_query(
        self,
        url_pattern,
        count=False,
        offset=0,
        limit=50,
        regexp=False,
        with_visit=False,
        visit_types=None,
        cur=None,
    ):
        """
        Method factorizing query creation for searching and counting origins.
        """
        cur = self._cursor(cur)

        if count:
            origin_cols = "COUNT(*)"
            order_clause = ""
        else:
            origin_cols = ",".join(self.origin_cols)
            order_clause = "ORDER BY id"

        if not regexp:
            operator = "ILIKE"
            query_params = [f"%{url_pattern}%"]
        else:
            operator = "~*"
            query_params = [url_pattern]

        query = f"""
            WITH filtered_origins AS (
                SELECT *
                FROM origin
                WHERE url {operator} %s
                {order_clause}
            )
            SELECT {origin_cols}
            FROM filtered_origins AS o
            """

        if with_visit or visit_types:
            visit_predicat = (
                """
                INNER JOIN origin_visit_status ovs USING (origin, visit)
                INNER JOIN snapshot ON ovs.snapshot=snapshot.id
                """
                if with_visit
                else ""
            )

            type_predicat = (
                f"AND ov.type=any(ARRAY{visit_types})" if visit_types else ""
            )

            query += f"""
                WHERE EXISTS (
                    SELECT 1
                    FROM origin_visit ov
                    {visit_predicat}
                    WHERE ov.origin=o.id {type_predicat}
                )
                """

        if not count:
            query += "OFFSET %s LIMIT %s"
            query_params.extend([offset, limit])

        cur.execute(query, query_params)

    def origin_search(
        self,
        url_pattern: str,
        offset: int = 0,
        limit: int = 50,
        regexp: bool = False,
        with_visit: bool = False,
        visit_types: Optional[List[str]] = None,
        cur=None,
    ):
        """Search for origins whose urls contain a provided string pattern
        or match a provided regular expression.
        The search is performed in a case insensitive way.

        Args:
            url_pattern: the string pattern to search for in origin urls
            offset: number of found origins to skip before returning
                results
            limit: the maximum number of found origins to return
            regexp: if True, consider the provided pattern as a regular
                expression and returns origins whose urls match it
            with_visit: if True, filter out origins with no visit

        """
        self._origin_query(
            url_pattern,
            offset=offset,
            limit=limit,
            regexp=regexp,
            with_visit=with_visit,
            visit_types=visit_types,
            cur=cur,
        )
        yield from cur

    def origin_count(self, url_pattern, regexp=False, with_visit=False, cur=None):
        """Count origins whose urls contain a provided string pattern
        or match a provided regular expression.
        The pattern search in origin urls is performed in a case insensitive
        way.

        Args:
            url_pattern (str): the string pattern to search for in origin urls
            regexp (bool): if True, consider the provided pattern as a regular
                expression and returns origins whose urls match it
            with_visit (bool): if True, filter out origins with no visit
        """
        self._origin_query(
            url_pattern, count=True, regexp=regexp, with_visit=with_visit, cur=cur
        )
        return cur.fetchone()[0]

    def origin_snapshot_get_all(self, origin_url: str, cur=None) -> Iterable[Sha1Git]:
        cur = self._cursor(cur)
        query = f"""\
        SELECT DISTINCT snapshot FROM origin_visit_status ovs
        INNER JOIN origin o ON o.id = ovs.origin
        WHERE o.url = '{origin_url}' and snapshot IS NOT NULL;
        """
        cur.execute(query)
        yield from map(lambda row: row[0], cur)

    ##########################
    # misc.
    ##########################

    object_find_by_sha1_git_cols = ["sha1_git", "type"]

    def object_find_by_sha1_git(self, ids, cur=None):
        cur = self._cursor(cur)

        yield from execute_values_generator(
            cur,
            """
            WITH t (sha1_git) AS (VALUES %s),
            known_objects as ((
                select
                  id as sha1_git,
                  'release'::object_type as type,
                  object_id
                from release r
                where exists (select 1 from t where t.sha1_git = r.id)
            ) union all (
                select
                  id as sha1_git,
                  'revision'::object_type as type,
                  object_id
                from revision r
                where exists (select 1 from t where t.sha1_git = r.id)
            ) union all (
                select
                  id as sha1_git,
                  'directory'::object_type as type,
                  object_id
                from directory d
                where exists (select 1 from t where t.sha1_git = d.id)
            ) union all (
                select
                  sha1_git as sha1_git,
                  'content'::object_type as type,
                  object_id
                from content c
                where exists (select 1 from t where t.sha1_git = c.sha1_git)
            ))
            select t.sha1_git as sha1_git, k.type
            from t
            left join known_objects k on t.sha1_git = k.sha1_git
            """,
            ((id,) for id in ids),
        )

    def stat_counters(self, cur=None):
        cur = self._cursor(cur)
        cur.execute("SELECT * FROM swh_stat_counters()")
        yield from cur

    ##########################
    # 'raw_extrinsic_metadata' table
    ##########################

    _raw_extrinsic_metadata_context_cols = [
        "origin",
        "visit",
        "snapshot",
        "release",
        "revision",
        "path",
        "directory",
    ]
    """The list of context columns for all artifact types."""

    _raw_extrinsic_metadata_insert_cols = [
        "id",
        "type",
        "target",
        "authority_id",
        "fetcher_id",
        "discovery_date",
        "format",
        "metadata",
        *_raw_extrinsic_metadata_context_cols,
    ]
    """List of columns of the raw_extrinsic_metadata table, used when writing
    metadata."""

    _raw_extrinsic_metadata_insert_query = f"""
        INSERT INTO raw_extrinsic_metadata
            ({', '.join(_raw_extrinsic_metadata_insert_cols)})
        VALUES ({', '.join('%s' for _ in _raw_extrinsic_metadata_insert_cols)})
        ON CONFLICT (id)
        DO NOTHING
    """

    raw_extrinsic_metadata_get_cols = [
        "raw_extrinsic_metadata.target",
        "raw_extrinsic_metadata.type",
        "discovery_date",
        "metadata_authority.type",
        "metadata_authority.url",
        "metadata_fetcher.id",
        "metadata_fetcher.name",
        "metadata_fetcher.version",
        *_raw_extrinsic_metadata_context_cols,
        "format",
        "raw_extrinsic_metadata.metadata",
    ]
    """List of columns of the raw_extrinsic_metadata, metadata_authority,
    and metadata_fetcher tables, used when reading object metadata."""

    _raw_extrinsic_metadata_select_query = f"""
        SELECT
            {', '.join(raw_extrinsic_metadata_get_cols)}
        FROM raw_extrinsic_metadata
        INNER JOIN metadata_authority
            ON (metadata_authority.id=authority_id)
        INNER JOIN metadata_fetcher ON (metadata_fetcher.id=fetcher_id)
    """

    def raw_extrinsic_metadata_add(
        self,
        id: bytes,
        type: str,
        target: str,
        discovery_date: datetime.datetime,
        authority_id: int,
        fetcher_id: int,
        format: str,
        metadata: bytes,
        origin: Optional[str],
        visit: Optional[int],
        snapshot: Optional[str],
        release: Optional[str],
        revision: Optional[str],
        path: Optional[bytes],
        directory: Optional[str],
        cur,
    ):
        query = self._raw_extrinsic_metadata_insert_query
        args: Dict[str, Any] = dict(
            id=id,
            type=type,
            target=target,
            authority_id=authority_id,
            fetcher_id=fetcher_id,
            discovery_date=discovery_date,
            format=format,
            metadata=metadata,
            origin=origin,
            visit=visit,
            snapshot=snapshot,
            release=release,
            revision=revision,
            path=path,
            directory=directory,
        )

        params = [args[col] for col in self._raw_extrinsic_metadata_insert_cols]

        cur.execute(query, params)

    def raw_extrinsic_metadata_get(
        self,
        target: str,
        authority_id: int,
        after_time: Optional[datetime.datetime],
        after_fetcher: Optional[int],
        limit: int,
        cur,
    ):
        query_parts = [self._raw_extrinsic_metadata_select_query]
        query_parts.append("WHERE raw_extrinsic_metadata.target=%s AND authority_id=%s")
        args = [target, authority_id]

        if after_fetcher is not None:
            assert after_time
            query_parts.append("AND (discovery_date, fetcher_id) > (%s, %s)")
            args.extend([after_time, after_fetcher])
        elif after_time is not None:
            query_parts.append("AND discovery_date > %s")
            args.append(after_time)

        query_parts.append("ORDER BY discovery_date, fetcher_id")

        if limit:
            query_parts.append("LIMIT %s")
            args.append(limit)

        cur.execute(" ".join(query_parts), args)
        yield from cur

    def raw_extrinsic_metadata_get_by_ids(self, ids: List[Sha1Git], cur=None):
        cur = self._cursor(cur)
        yield from execute_values_generator(
            cur,
            self._raw_extrinsic_metadata_select_query
            + "INNER JOIN (VALUES %s) AS t(id) ON t.id = raw_extrinsic_metadata.id",
            [(id_,) for id_ in ids],
        )

    def raw_extrinsic_metadata_get_authorities(self, id: str, cur=None):
        cur = self._cursor(cur)
        cur.execute(
            """
            SELECT
                DISTINCT metadata_authority.type, metadata_authority.url
            FROM raw_extrinsic_metadata
            INNER JOIN metadata_authority
                ON (metadata_authority.id=authority_id)
            WHERE raw_extrinsic_metadata.target = %s
            """,
            (id,),
        )
        yield from cur

    ##########################
    # 'metadata_fetcher' and 'metadata_authority' tables
    ##########################

    metadata_fetcher_cols = ["name", "version"]

    def metadata_fetcher_add(self, name: str, version: str, cur=None) -> None:
        cur = self._cursor(cur)
        cur.execute(
            "INSERT INTO metadata_fetcher (name, version) "
            "VALUES (%s, %s) ON CONFLICT DO NOTHING",
            (name, version),
        )

    def metadata_fetcher_get(self, name: str, version: str, cur=None):
        cur = self._cursor(cur)
        cur.execute(
            f"SELECT {', '.join(self.metadata_fetcher_cols)} "
            f"FROM metadata_fetcher "
            f"WHERE name=%s AND version=%s",
            (name, version),
        )
        return cur.fetchone()

    def metadata_fetcher_get_id(
        self, name: str, version: str, cur=None
    ) -> Optional[int]:
        cur = self._cursor(cur)
        cur.execute(
            "SELECT id FROM metadata_fetcher WHERE name=%s AND version=%s",
            (name, version),
        )
        row = cur.fetchone()
        if row:
            return row[0]
        else:
            return None

    metadata_authority_cols = ["type", "url"]

    def metadata_authority_add(self, type: str, url: str, cur=None) -> None:
        cur = self._cursor(cur)
        cur.execute(
            "INSERT INTO metadata_authority (type, url) "
            "VALUES (%s, %s) ON CONFLICT DO NOTHING",
            (type, url),
        )

    def metadata_authority_get(self, type: str, url: str, cur=None):
        cur = self._cursor(cur)
        cur.execute(
            f"SELECT {', '.join(self.metadata_authority_cols)} "
            f"FROM metadata_authority "
            f"WHERE type=%s AND url=%s",
            (type, url),
        )
        return cur.fetchone()

    def metadata_authority_get_id(self, type: str, url: str, cur=None) -> Optional[int]:
        cur = self._cursor(cur)
        cur.execute(
            "SELECT id FROM metadata_authority WHERE type=%s AND url=%s", (type, url)
        )
        row = cur.fetchone()
        if row:
            return row[0]
        else:
            return None

    # 'object_references' table

    _object_references_cols = ("source_type", "source", "target_type", "target")

    def object_references_get(
        self, target_type: str, target: bytes, limit: int, cur=None
    ):
        cur = self._cursor(cur)
        cur.execute(
            """SELECT %s
            FROM object_references
            WHERE target_type = %%s and target=%%s
            LIMIT %%s"""
            % (", ".join(self._object_references_cols)),
            (target_type, target, limit),
        )
        return [dict(zip(self._object_references_cols, row)) for row in cur.fetchall()]

    def object_references_add(self, reference_rows, cur=None) -> None:
        cols = ", ".join(self._object_references_cols)
        cur = self._cursor(cur)
        execute_values(
            cur,
            f"""INSERT INTO object_references ({cols})
            VALUES %s ON CONFLICT (insertion_date, {cols}) DO NOTHING""",
            reference_rows,
        )

    def object_references_create_partition(
        self, year: int, week: int, cur=None
    ) -> Tuple[datetime.date, datetime.date]:
        """Create the partition of the object_references table for the given ISO
        ``year`` and ``week``."""
        # This date is guaranteed to be in week 1 by the ISO standard
        in_week1 = datetime.date(year=year, month=1, day=4)
        monday_of_week1 = in_week1 + datetime.timedelta(days=-in_week1.weekday())

        monday = monday_of_week1 + datetime.timedelta(weeks=week - 1)
        next_monday = monday + datetime.timedelta(days=7)

        cur = self._cursor(cur)
        cur.execute(
            """
            CREATE TABLE IF NOT EXISTS object_references_%04dw%02d
            PARTITION OF object_references
            FOR VALUES FROM (%%s) TO (%%s)"""
            % (year, week),
            (monday.isoformat(), next_monday.isoformat()),
        )

        return (monday, next_monday)

    def object_references_drop_partition(self, year: int, week: int, cur=None) -> None:
        """Delete the partition of the object_references table for the given ISO
        ``year`` and ``week``."""
        cur = self._cursor(cur)
        cur.execute("DROP TABLE object_references_%04dw%02d" % (year, week))

    def object_references_list_partitions(
        self, cur=None
    ) -> List[ObjectReferencesPartition]:
        """List existing partitions of the object_references table, ordered from
        oldest to the most recent."""
        cur = self._cursor(cur)
        cur.execute(
            """SELECT relname, pg_get_expr(relpartbound, oid)
                 FROM pg_partition_tree('object_references') pt
                      INNER JOIN pg_class ON relid = pg_class.oid
                WHERE isleaf = true
                ORDER BY relname"""
        )
        name_re = re.compile(r"^object_references_([0-9]+)w([0-9]+)$")
        bounds_re = re.compile(r"^FOR VALUES FROM \('([0-9-]+)'\) TO \('([0-9-]+)'\)$")
        partitions = []
        for row in cur:
            name_m = name_re.match(row[0])
            assert name_m is not None
            bounds_m = bounds_re.match(row[1])
            assert bounds_m is not None
            partitions.append(
                ObjectReferencesPartition(
                    table_name=row[0],
                    year=int(name_m[1]),
                    week=int(name_m[2]),
                    start=datetime.datetime.fromisoformat(bounds_m[1]),
                    end=datetime.datetime.fromisoformat(bounds_m[2]),
                )
            )
        return partitions

    #################
    # object deletion
    #################

    def object_delete(
        self, object_rows: List[Tuple[str, bytes]], cur=None
    ) -> Dict[str, int]:
        result = {}
        cur = self._cursor(cur)
        cur.execute(
            """
            CREATE TEMPORARY TABLE objects_to_remove (
                type extended_object_type NOT NULL,
                id BYTEA NOT NULL
            ) ON COMMIT DROP
            """
        )
        execute_values(
            cur,
            """INSERT INTO objects_to_remove (type, id)
               VALUES %s""",
            object_rows,
        )
        # Let’s handle raw extrinsic metadata first as they’ll
        # be referencing other objects
        cur.execute(
            """SELECT COUNT(emd.id), emd.type AS target_type
                 FROM raw_extrinsic_metadata emd
                WHERE emd.id IN (SELECT id
                                   FROM objects_to_remove
                                  WHERE type = 'raw_extrinsic_metadata')
                GROUP BY type"""
        )
        for count, target_type in cur:
            result[
                f"{ExtendedObjectType[target_type.upper()].value}_metadata:delete"
            ] = count
        cur.execute(
            """DELETE FROM raw_extrinsic_metadata emd
                WHERE emd.id IN (SELECT id
                                   FROM objects_to_remove
                                  WHERE type = 'raw_extrinsic_metadata')"""
        )
        # We need to remove lines from `origin_visit_status`,
        # `origin_visit` and `origin`.
        cur.execute(
            """CREATE TEMPORARY TABLE origins_to_remove
                 ON COMMIT DROP
                 AS
                   SELECT origin.id FROM origin
                    INNER JOIN objects_to_remove otr
                       ON DIGEST(url, 'sha1') = otr.id and otr.type = 'origin'"""
        )
        cur.execute(
            """DELETE FROM origin_visit_status ovs
                WHERE ovs.origin IN (SELECT id FROM origins_to_remove)"""
        )
        result["origin_visit_status:delete"] = cur.rowcount
        cur.execute(
            """DELETE FROM origin_visit ov
                WHERE ov.origin IN (SELECT id FROM origins_to_remove)"""
        )
        result["origin_visit:delete"] = cur.rowcount
        cur.execute(
            """DELETE FROM origin
                WHERE origin.id IN (SELECT id FROM origins_to_remove)"""
        )
        result["origin:delete"] = cur.rowcount
        # We must not remove entries for snapshot_branch, they are shared across
        # multiple snapshots, and we don't keep a reverse index (to know if the
        # shared branches are still used by another snapshot).
        cur.execute(
            """CREATE TEMPORARY TABLE snapshots_to_remove
                  ON COMMIT DROP
                  AS
                    SELECT object_id
                      FROM snapshot s
                     INNER JOIN objects_to_remove otr
                        ON s.id = otr.id AND otr.type = 'snapshot'"""
        )
        cur.execute(
            """DELETE FROM snapshot_branches
                WHERE snapshot_branches.snapshot_id IN
                  (SELECT object_id FROM snapshots_to_remove)"""
        )
        cur.execute(
            """DELETE FROM snapshot
                WHERE snapshot.object_id IN (SELECT object_id
                                               FROM snapshots_to_remove)"""
        )
        result["snapshot:delete"] = cur.rowcount
        cur.execute(
            """DELETE FROM release
                WHERE release.id IN (SELECT id
                                       FROM objects_to_remove
                                      WHERE type = 'release')"""
        )
        result["release:delete"] = cur.rowcount
        cur.execute(
            """DELETE FROM revision_history
                WHERE revision_history.id IN (SELECT id
                                                FROM objects_to_remove
                                               WHERE type = 'revision')"""
        )
        cur.execute(
            """DELETE FROM revision
                WHERE revision.id IN (SELECT id
                                        FROM objects_to_remove
                                       WHERE type = 'revision')"""
        )
        result["revision:delete"] = cur.rowcount
        # We do not remove anything from `directory_entry_dir`,
        # `directory_entry_file`, `directory_entry_rev`: these entries are
        # shared across directories and we don't have (or don’t want to keep) an
        # index to know which directory uses what entry.
        cur.execute(
            """DELETE FROM directory
                WHERE directory.id IN (SELECT id
                                         FROM objects_to_remove
                                        WHERE type = 'directory')"""
        )
        result["directory:delete"] = cur.rowcount
        cur.execute(
            """DELETE FROM skipped_content
                WHERE skipped_content.sha1_git IN (
                    SELECT id
                      FROM objects_to_remove
                     WHERE type = 'content')"""
        )
        result["skipped_content:delete"] = cur.rowcount
        cur.execute(
            """DELETE FROM content
                WHERE content.sha1_git IN (
                    SELECT id
                      FROM objects_to_remove
                     WHERE type = 'content')"""
        )
        result["content:delete"] = cur.rowcount
        # We are not an objstorage
        result["content:delete:bytes"] = 0
        return result



# Copyright (C) 2018-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import dataclasses

import pytest

from swh.storage.cassandra.model import BaseRow
from swh.storage.in_memory import Table
from swh.storage.tests.storage_tests import (
    TestStorageGeneratedData as _TestStorageGeneratedData,
)
from swh.storage.tests.storage_tests import TestStorage as _TestStorage

# tests are executed using imported classes (TestStorage and
# TestStorageGeneratedData) using overloaded swh_storage fixture
# below


@pytest.fixture
def swh_storage_backend_config():
    yield {
        "cls": "memory",
        "journal_writer": {
            "cls": "memory",
        },
    }


@dataclasses.dataclass
class Row(BaseRow):
    PARTITION_KEY = ("col1", "col2")
    CLUSTERING_KEY = ("col3", "col4")

    col1: str
    col2: str
    col3: str
    col4: str
    col5: str
    col6: int


def test_table_keys():
    table = Table(Row)

    primary_key = ("foo", "bar", "baz", "qux")
    partition_key = ("foo", "bar")
    clustering_key = ("baz", "qux")

    row = Row(col1="foo", col2="bar", col3="baz", col4="qux", col5="quux", col6=4)
    assert table.partition_key(row) == partition_key
    assert table.clustering_key(row) == clustering_key
    assert table.primary_key(row) == primary_key

    assert table.primary_key_from_dict(row.to_dict()) == primary_key
    assert table.split_primary_key(primary_key) == (partition_key, clustering_key)


def test_table():
    table = Table(Row)

    row1 = Row(col1="foo", col2="bar", col3="baz", col4="qux", col5="quux", col6=4)
    row2 = Row(col1="foo", col2="bar", col3="baz", col4="qux2", col5="quux", col6=4)
    row3 = Row(col1="foo", col2="bar", col3="baz", col4="qux1", col5="quux", col6=4)
    row4 = Row(col1="foo", col2="bar2", col3="baz", col4="qux1", col5="quux", col6=4)
    partition_key = ("foo", "bar")
    partition_key4 = ("foo", "bar2")
    primary_key1 = ("foo", "bar", "baz", "qux")
    primary_key2 = ("foo", "bar", "baz", "qux2")
    primary_key3 = ("foo", "bar", "baz", "qux1")
    primary_key4 = ("foo", "bar2", "baz", "qux1")

    table.insert(row1)
    table.insert(row2)
    table.insert(row3)
    table.insert(row4)

    assert table.get_from_primary_key(primary_key1) == row1
    assert table.get_from_primary_key(primary_key2) == row2
    assert table.get_from_primary_key(primary_key3) == row3
    assert table.get_from_primary_key(primary_key4) == row4

    # order matters
    assert list(table.get_from_token(table.token(partition_key))) == [row1, row3, row2]

    # order matters
    assert list(table.get_from_partition_key(partition_key)) == [row1, row3, row2]

    assert list(table.get_from_partition_key(partition_key4)) == [row4]

    all_rows = list(table.iter_all())
    assert len(all_rows) == 4
    for row in (row1, row2, row3, row4):
        assert (table.primary_key(row), row) in all_rows


class TestInMemoryStorage(_TestStorage):
    @pytest.mark.skip(
        'The "person" table of the pgsql is a legacy thing, and not '
        "supported by the cassandra backend."
    )
    def test_person_fullname_unicity(self):
        pass

    @pytest.mark.skip("content_update is not yet implemented for Cassandra")
    def test_content_update(self):
        pass

    @pytest.mark.skip("Not supported by Cassandra")
    def test_origin_count(self):
        pass

    @pytest.mark.skip("in-memory backend has no timeout")
    def test_querytimeout(self):
        pass


class TestInMemoryStorageGeneratedData(_TestStorageGeneratedData):
    @pytest.mark.skip("Not supported by Cassandra")
    def test_origin_count(self):
        pass

    @pytest.mark.skip("Not supported by Cassandra")
    def test_origin_count_with_visit_no_visits(self):
        pass

    @pytest.mark.skip("Not supported by Cassandra")
    def test_origin_count_with_visit_with_visits_and_snapshot(self):
        pass

    @pytest.mark.skip("Not supported by Cassandra")
    def test_origin_count_with_visit_with_visits_no_snapshot(self):
        pass



# Copyright (C) 2024 The Software Heritage developers

# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import enum
import logging
from typing import Dict, List, Optional, Tuple
from urllib.parse import urlparse
from uuid import UUID

import attr
import psycopg2.errors
from psycopg2.extras import execute_values

from swh.core.db import BaseDb
from swh.core.statsd import statsd
from swh.storage.proxies.masking.db import DuplicateRequest, RequestNotFound

METRIC_QUERY_TOTAL = "swh_storage_blocking_queried_total"
METRIC_BLOCKED_TOTAL = "swh_storage_blocking_blocked_total"
KNOWN_SUFFIXES = ("git", "svn", "hg", "cvs", "CVS")

logger = logging.getLogger(__name__)


class BlockingState(enum.Enum):
    """Value recording "how much" an url associated to a blocking request is blocked"""

    NON_BLOCKED = enum.auto()
    """The origin url can be ingested/updated"""
    DECISION_PENDING = enum.auto()
    """Ingestion from origin url is temporarily blocked until the request is reviewed"""
    BLOCKED = enum.auto()
    """Ingestion from origin url is permanently blocked"""


@attr.s
class BlockingStatus:
    """Return value when requesting if an origin url ingestion is blocked"""

    state = attr.ib(type=BlockingState)
    request = attr.ib(type=UUID)


@attr.s
class BlockingRequest:
    """A request for blocking a set of origins from being ingested"""

    id = attr.ib(type=UUID)
    """Unique id for the request (will be returned to requesting clients)"""
    slug = attr.ib(type=str)
    """Unique, human-readable id for the request (for administrative interactions)"""
    date = attr.ib(type=datetime.datetime)
    """Date the request was received"""
    reason = attr.ib(type=str)
    """Why the request was made"""  # TODO: should this be stored here?


@attr.s
class RequestHistory:
    request = attr.ib(type=UUID)
    """id of the blocking request"""
    date = attr.ib(type=datetime.datetime)
    """Date the history entry has been added"""
    message = attr.ib(type=str)
    """Free-form history information (e.g. "policy decision made")"""


@attr.s
class BlockingLogEntry:
    url = attr.ib(type=str)
    """origin url that have been blocked"""
    url_match = attr.ib(type=str)
    """url matching pattern that caused the blocking of the origin url"""
    request = attr.ib(type=UUID)
    """id of the blocking request"""
    date = attr.ib(type=datetime.datetime)
    """Date the blocking event occurred"""
    state = attr.ib(type=BlockingState)
    """Blocking state responsible for the blocking event"""


@attr.s
class BlockedOrigin:
    request_slug = attr.ib(type=str)
    url_pattern = attr.ib(type=str)
    state = attr.ib(type=BlockingState)


class BlockingDb(BaseDb):
    current_version = 1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.conn.autocommit = True


def get_urls_to_check(url: str) -> Tuple[List[str], List[str]]:
    """Get the entries to check in the database for the given `url`, in order.

    Exact matching is done on the following strings, in order:
     - the url with any trailing slashes removed (the so-called "trimmed url");
     - the url passed exactly;
     - if the trimmed url ends with a dot and one of the
       :const:`KNOWN_SUFFIXES`, the url with this suffix stripped.

    The prefix matching is done by splitting the path part of the URL on
    slashes, and successively removing the last elements.

    Returns:
      A tuple with a list of exact matches, and a list of prefix matches

    """

    # Generate exact strings to match against
    exact_matches = [url]

    if url.endswith("/"):
        trimmed_url = url.rstrip("/")
        exact_matches.insert(0, trimmed_url)
    else:
        trimmed_url = url

    if "." in trimmed_url:
        stripped_url, suffix = trimmed_url.rsplit(".", 1)
        if suffix in KNOWN_SUFFIXES:
            exact_matches.append(stripped_url)

    # Generate prefixes to match against
    parsed_url = urlparse(trimmed_url)
    base_url = f"{parsed_url.scheme}://{parsed_url.netloc}"

    prefix_matches = []

    split_path = parsed_url.path.lstrip("/").split("/")
    for i in range(1, len(split_path) + 1):
        prefix_matches.append("/".join([base_url] + split_path[: len(split_path) - i]))

    return exact_matches, prefix_matches


class BlockingAdmin(BlockingDb):
    def create_request(self, slug: str, reason: str) -> BlockingRequest:
        """Record a new blocking request

        Arguments:
          slug: human-readable unique identifier for the request
          reason: free-form text recording why the request was made

        Raises:
          DuplicateRequest when the slug already exists
        """
        cur = self.cursor()

        try:
            cur.execute(
                """
                INSERT INTO blocking_request (slug, reason)
                VALUES (%s, %s) RETURNING id, date
                """,
                (slug, reason),
            )
        except psycopg2.errors.UniqueViolation:
            raise DuplicateRequest(slug)

        res = cur.fetchone()
        assert res is not None, "PostgreSQL returned an inconsistent result"
        id, date = res
        return BlockingRequest(id=id, date=date, slug=slug, reason=reason)

    def find_request(self, slug: str) -> Optional[BlockingRequest]:
        """Find a blocking request using its slug

        Returns: :const:`None` if a request with the given slug doesn't exist
        """
        cur = self.cursor()
        cur.execute(
            """
            SELECT id, slug, date, reason
            FROM blocking_request
            WHERE slug = %s
            """,
            (slug,),
        )

        res = cur.fetchone()
        if not res:
            return None
        id, slug, date, reason = res
        return BlockingRequest(id=id, date=date, slug=slug, reason=reason)

    def find_request_by_id(self, id: UUID) -> Optional[BlockingRequest]:
        """Find a blocking request using its id

        Returns: :const:`None` if a request with the given request doesn't exist
        """
        cur = self.cursor()
        cur.execute(
            """
            SELECT id, slug, date, reason
            FROM blocking_request
            WHERE id = %s
            """,
            (id,),
        )

        res = cur.fetchone()
        if not res:
            return None
        id, slug, date, reason = res
        return BlockingRequest(id=id, date=date, slug=slug, reason=reason)

    def get_requests(
        self, include_cleared_requests: bool = False
    ) -> List[Tuple[BlockingRequest, int]]:
        """Get known requests

        Args:
            include_cleared_requests: also include requests with no associated
            blocking states
        """
        cur = self.cursor()

        query = """SELECT id, slug, date, reason, COUNT(url_match) AS blocking_count
                     FROM blocking_request
                     LEFT JOIN blocked_origin ON (request = id)
                     GROUP BY id"""
        if not include_cleared_requests:
            query += "  HAVING COUNT(url_match) > 0"
        query += "  ORDER BY date DESC"
        cur.execute(query)
        result = []
        for id, slug, date, reason, block_count in cur:
            result.append(
                (
                    BlockingRequest(id=id, slug=slug, date=date, reason=reason),
                    block_count,
                )
            )
        return result

    def set_origins_state(
        self, request_id: UUID, new_state: BlockingState, urls: List[str]
    ):
        """Within the request with the given id, record the state of the given
        objects as ``new_state``.

        This creates entries or updates them as appropriate.

        Raises: :exc:`RequestNotFound` if the request is not found.
        """
        cur = self.cursor()

        cur.execute("SELECT 1 FROM blocking_request WHERE id = %s", (request_id,))
        if cur.fetchone() is None:
            raise RequestNotFound(request_id)

        execute_values(
            cur,
            """
            INSERT INTO blocked_origin (url_match, request, state)
            VALUES %s
            ON CONFLICT (url_match, request)
            DO UPDATE SET state = EXCLUDED.state
            """,
            (
                (
                    url,
                    request_id,
                    new_state.name.lower(),
                )
                for url in urls
            ),
        )

    def get_states_for_request(self, request_id: UUID) -> Dict[str, BlockingState]:
        """Get the state of urls associated with the given request.

        Raises :exc:`RequestNotFound` if the request is not found.
        """

        cur = self.cursor()

        cur.execute("SELECT 1 FROM blocking_request WHERE id = %s", (request_id,))
        if cur.fetchone() is None:
            raise RequestNotFound(request_id)

        result = {}
        cur.execute(
            """SELECT url_match, state
                 FROM blocked_origin
                WHERE request = %s""",
            (request_id,),
        )
        for url, state in cur:
            result[url] = BlockingState[state.upper()]
        return result

    def find_blocking_states(self, urls: List[str]) -> List[BlockedOrigin]:
        """Lookup the blocking state and associated requests for the given urls
        (exact match)."""

        cur = self.cursor()
        result = []
        for (
            url,
            request_slug,
            state,
        ) in psycopg2.extras.execute_values(
            cur,
            """SELECT url_match, slug, state
                 FROM blocked_origin
                INNER JOIN (VALUES %s) v(url_match)
                USING (url_match)
                 LEFT JOIN blocking_request ON (blocking_request.id = request)
                ORDER BY url_match, blocking_request.date DESC
            """,
            ((url,) for url in urls),
            fetch=True,
        ):
            result.append(
                BlockedOrigin(
                    request_slug=request_slug,
                    url_pattern=url,
                    state=BlockingState[state.upper()],
                )
            )
        return result

    def delete_blocking_states(self, request_id: UUID) -> None:
        """Remove all blocking states for the given request.

        Raises: :exc:`RequestNotFound` if the request is not found.
        """

        cur = self.cursor()
        cur.execute("SELECT 1 FROM blocking_request WHERE id = %s", (request_id,))
        if cur.fetchone() is None:
            raise RequestNotFound(request_id)

        cur.execute("DELETE FROM blocked_origin WHERE request = %s", (request_id,))

    def record_history(self, request_id: UUID, message: str) -> RequestHistory:
        """Add an entry to the history of the given request.

        Raises: :exc:`RequestNotFound` if the request is not found.
        """
        cur = self.cursor()
        try:
            cur.execute(
                """
                INSERT INTO blocking_request_history (request, message)
                VALUES (%s, %s) RETURNING date
                """,
                (request_id, message),
            )
        except psycopg2.errors.ForeignKeyViolation:
            raise RequestNotFound(request_id)

        res = cur.fetchone()
        assert res is not None, "PostgreSQL returned an inconsistent result"

        return RequestHistory(request=request_id, date=res[0], message=message)

    def get_history(self, request_id: UUID) -> List[RequestHistory]:
        """Get the history of a given request.

        Raises: :exc:`RequestNotFound` if the request if not found.
        """
        cur = self.cursor()

        cur.execute("SELECT 1 FROM blocking_request WHERE id = %s", (request_id,))
        if cur.fetchone() is None:
            raise RequestNotFound(request_id)

        cur.execute(
            """SELECT date, message
                    FROM blocking_request_history
                WHERE request = %s
                ORDER BY date DESC""",
            (request_id,),
        )
        records = []
        for date, message in cur:
            records.append(
                RequestHistory(request=request_id, date=date, message=message)
            )
        return records

    def get_log(
        self, request_id: Optional[UUID] = None, url: Optional[str] = None
    ) -> List[BlockingLogEntry]:
        cur = self.cursor()
        where = []
        args = []
        if request_id:
            where.append("request = %s")
            args.append(str(request_id))
        if url:
            where.append("url = %s")
            args.append(url)
        if where:
            condition = "WHERE " + " AND ".join(where)
        else:
            condition = ""
        cur.execute(
            f"""SELECT date, url, url_match, request, state
                    FROM blocked_origin_log
                    {condition}
                    ORDER BY date DESC""",
            args,
        )
        records = []
        for db_date, db_url, db_url_match, db_request_id, db_state in cur:
            records.append(
                BlockingLogEntry(
                    url=db_url,
                    url_match=db_url_match,
                    request=db_request_id,
                    date=db_date,
                    state=BlockingState[db_state.upper()],
                )
            )
        return records


class BlockingQuery(BlockingDb):
    def origins_are_blocked(
        self, urls: List[str], all_statuses=False
    ) -> Dict[str, BlockingStatus]:
        """Return the blocking status for eeach origin url given in urls

        If all_statuses is False, do not return urls whose blocking status is
        defined as NON_BLOCKING (so only return actually blocked urls).
        Otherwise, return all matching blocking status.

        """
        ret = {}
        for url in urls:
            status = self.origin_is_blocked(url)
            if status and (all_statuses or status.state != BlockingState.NON_BLOCKED):
                ret[url] = status
        return ret

    def origin_is_blocked(self, url: str) -> Optional[BlockingStatus]:
        """Checks if the origin URL should be blocked.

        If the given url matches a set of registered blocking rules, return the
        most appropriate one. Otherwise, return None.

        Log the blocking event in the database (log only a matching events).
        """
        logging.debug("url: %s", url)
        statsd.increment(METRIC_QUERY_TOTAL, 1)

        exact_matches, prefix_matches = get_urls_to_check(url)

        cur = self.cursor()
        psycopg2.extras.execute_values(
            cur,
            """
            SELECT url_match, request, state
            FROM blocked_origin
            INNER JOIN (VALUES %s) v(url_match)
            USING (url_match)
            ORDER BY char_length(url_match) DESC LIMIT 1
            """,
            ((_url,) for _url in exact_matches + prefix_matches),
        )
        row = cur.fetchone()

        if row:
            url_match, request_id, state = row
            status = BlockingStatus(
                state=BlockingState[state.upper()], request=request_id
            )
            logger.debug("Matching status for %s: %s", url_match, status)
            statsd.increment(METRIC_BLOCKED_TOTAL, 1)
            # log the event; even a NON_BLOCKED decision is logged
            cur.execute(
                """
                INSERT INTO blocked_origin_log (url, url_match, request, state)
                VALUES (%s, %s, %s, %s)
                """,
                (url, url_match, status.request, status.state.name.lower()),
            )
            return status
        return None



# Copyright (C) 2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from typing import Dict, List, Mapping

from swh.model.model import Directory, OriginVisitStatus, Release, Revision, Snapshot
from swh.storage import get_storage
from swh.storage.interface import ObjectReference, StorageInterface


class RecordReferencesProxyStorage:
    """Automatically store object references when adding objects that have them"""

    def __init__(self, storage: Mapping):
        self.storage: StorageInterface = get_storage(**storage)

    def __getattr__(self, key: str):
        if key == "storage":
            raise AttributeError(key)
        return getattr(self.storage, key)

    def directory_add(self, directories: List[Directory]) -> Dict[str, int]:
        object_references = [
            ObjectReference(
                source=directory.swhid().to_extended(),
                target=entry.swhid().to_extended(),
            )
            for directory in directories
            for entry in directory.entries
        ]
        stats = self.storage.object_references_add(object_references)
        stats.update(self.storage.directory_add(directories))
        return stats

    def revision_add(self, revisions: List[Revision]) -> Dict[str, int]:
        object_references = []
        for revision in revisions:
            object_references.append(
                ObjectReference(
                    source=revision.swhid().to_extended(),
                    target=revision.directory_swhid().to_extended(),
                )
            )
            for swhid in revision.parent_swhids():
                object_references.append(
                    ObjectReference(
                        source=revision.swhid().to_extended(),
                        target=swhid.to_extended(),
                    )
                )

        stats = self.storage.object_references_add(object_references)
        stats.update(self.storage.revision_add(revisions))
        return stats

    def release_add(self, releases: List[Release]) -> Dict[str, int]:
        object_references = []
        for release in releases:
            target_swhid = release.target_swhid()
            if target_swhid:
                object_references.append(
                    ObjectReference(
                        source=release.swhid().to_extended(),
                        target=target_swhid.to_extended(),
                    )
                )
        stats = self.storage.object_references_add(object_references)
        stats.update(self.storage.release_add(releases))
        return stats

    def snapshot_add(self, snapshots: List[Snapshot]) -> Dict[str, int]:
        object_references = []
        for snapshot in snapshots:
            snapshot_swhid = snapshot.swhid().to_extended()
            for branch in snapshot.branches.values():
                if not branch:
                    continue
                target_swhid = branch.swhid()
                if target_swhid:
                    object_references.append(
                        ObjectReference(
                            source=snapshot_swhid,
                            target=target_swhid.to_extended(),
                        )
                    )

        stats = self.storage.object_references_add(object_references)
        stats.update(self.storage.snapshot_add(snapshots))
        return stats

    def origin_visit_status_add(
        self,
        visit_statuses: List[OriginVisitStatus],
    ) -> Dict[str, int]:
        object_references = []
        for visit_status in visit_statuses:
            snapshot_swhid = visit_status.snapshot_swhid()
            if snapshot_swhid:
                object_references.append(
                    ObjectReference(
                        source=visit_status.origin_swhid(),
                        target=snapshot_swhid.to_extended(),
                    )
                )

        stats = self.storage.object_references_add(object_references)
        stats.update(self.storage.origin_visit_status_add(visit_statuses))
        return stats



# Copyright (C) 2015-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from uuid import uuid4

import psycopg2.errors
import pytest

from swh.core.api import RemoteException, TransientRemoteException
from swh.model.model import Origin
from swh.model.swhids import ExtendedSWHID
import swh.storage
from swh.storage import get_storage
import swh.storage.api.server as server
from swh.storage.exc import BlockedOriginException, MaskedObjectException
from swh.storage.proxies.blocking.db import BlockingState, BlockingStatus
from swh.storage.proxies.masking.db import MaskedState, MaskedStatus
from swh.storage.tests.storage_tests import (
    TestStorageGeneratedData as _TestStorageGeneratedData,
)
from swh.storage.tests.storage_tests import TestStorage as _TestStorage

# tests are executed using imported classes (TestStorage and
# TestStorageGeneratedData) using overloaded swh_storage fixture
# below


@pytest.fixture
def app_server():
    server.storage = swh.storage.get_storage(
        cls="memory", journal_writer={"cls": "memory"}
    )
    yield server


@pytest.fixture
def app(app_server):
    return app_server.app


@pytest.fixture
def swh_rpc_client_class():
    def storage_factory(**kwargs):
        storage_config = {
            "cls": "remote",
            **kwargs,
        }
        return get_storage(**storage_config)

    return storage_factory


@pytest.fixture
def swh_storage(swh_rpc_client, app_server):
    # This version of the swh_storage fixture uses the swh_rpc_client fixture
    # to instantiate a RemoteStorage (see swh_rpc_client_class above) that
    # proxies, via the swh.core RPC mechanism, the local (in memory) storage
    # configured in the app_server fixture above.
    #
    # Also note that, for the sake of
    # making it easier to write tests, the in-memory journal writer of the
    # in-memory backend storage is attached to the RemoteStorage as its
    # journal_writer attribute.
    storage = swh_rpc_client

    journal_writer = getattr(storage, "journal_writer", None)
    storage.journal_writer = app_server.storage.journal_writer
    storage.objstorage = app_server.storage.objstorage
    yield storage
    storage.journal_writer = journal_writer


@pytest.fixture
def swh_storage_backend(app_server, swh_storage):
    return app_server.storage


class TestStorageApi(_TestStorage):
    @pytest.mark.skip(
        'The "person" table of the pgsql is a legacy thing, and not '
        "supported by the cassandra backend."
    )
    def test_person_fullname_unicity(self):
        pass

    @pytest.mark.skip("content_update is not yet implemented for Cassandra")
    def test_content_update(self):
        pass

    @pytest.mark.skip("Not supported by Cassandra")
    def test_origin_count(self):
        pass

    @pytest.mark.skip("in-memory backend has no timeout")
    def test_querytimeout(self):
        pass

    def test_exception(self, app_server, swh_storage, mocker):
        """Checks the client re-raises unknown exceptions as a :exc:`RemoteException`"""
        assert swh_storage.revision_get(["\x01" * 20]) == [None]
        mocker.patch.object(
            app_server.storage._cql_runner,
            "revision_get",
            side_effect=ValueError("crash"),
        )
        with pytest.raises(RemoteException) as e:
            swh_storage.revision_get(["\x01" * 20])
        assert not isinstance(e, TransientRemoteException)

    def test_operationalerror_exception(self, app_server, swh_storage, mocker):
        """Checks the client re-raises as a :exc:`TransientRemoteException`
        rather than the base :exc:`RemoteException`; so the retrying proxy
        retries for longer."""
        assert swh_storage.revision_get(["\x01" * 20]) == [None]
        mocker.patch.object(
            app_server.storage._cql_runner,
            "revision_get",
            side_effect=psycopg2.errors.AdminShutdown("cluster is shutting down"),
        )
        with pytest.raises(RemoteException) as excinfo:
            swh_storage.revision_get(["\x01" * 20])
        assert isinstance(excinfo.value, TransientRemoteException)

    def test_querycancelled_exception(self, app_server, swh_storage, mocker):
        """Checks the client re-raises as a :exc:`TransientRemoteException`
        rather than the base :exc:`RemoteException`; so the retrying proxy
        retries for longer."""
        assert swh_storage.revision_get(["\x01" * 20]) == [None]
        mocker.patch.object(
            app_server.storage._cql_runner,
            "revision_get",
            side_effect=psycopg2.errors.QueryCanceled("too big!"),
        )
        with pytest.raises(RemoteException) as excinfo:
            swh_storage.revision_get(["\x01" * 20])
        assert not isinstance(excinfo.value, TransientRemoteException)

    def test_masked_object_exception(self, app_server, swh_storage, mocker):
        """Checks the client re-raises masking proxy exceptions"""
        assert swh_storage.revision_get(["\x01" * 20]) == [None]
        masked = {
            ExtendedSWHID.from_string("swh:1:rev:" + ("01" * 20)): [
                MaskedStatus(MaskedState.DECISION_PENDING, request=uuid4())
            ]
        }
        mocker.patch.object(
            app_server.storage._cql_runner,
            "revision_get",
            side_effect=MaskedObjectException(masked),
        )
        with pytest.raises(MaskedObjectException) as e:
            swh_storage.revision_get(["\x01" * 20])
        assert e.value.masked == masked

    def test_blocked_origin_exception(self, app_server, swh_storage, mocker):
        """Checks the client re-raises masking proxy exceptions"""
        assert swh_storage.origin_get(["https://example.com"]) == [None]
        blocked = {
            "https://example.com": BlockingStatus(
                BlockingState.DECISION_PENDING, request=uuid4()
            )
        }
        mocker.patch.object(
            app_server.storage._cql_runner,
            "origin_add_one",
            side_effect=BlockedOriginException(blocked),
        )
        with pytest.raises(BlockedOriginException) as e:
            swh_storage.origin_add([Origin(url="https://example.com")])
        assert e.value.blocked == blocked


class TestStorageApiGeneratedData(_TestStorageGeneratedData):
    @pytest.mark.skip("Not supported by Cassandra")
    def test_origin_count(self):
        pass

    @pytest.mark.skip("Not supported by Cassandra")
    def test_origin_count_with_visit_no_visits(self):
        pass

    @pytest.mark.skip("Not supported by Cassandra")
    def test_origin_count_with_visit_with_visits_and_snapshot(self):
        pass

    @pytest.mark.skip("Not supported by Cassandra")
    def test_origin_count_with_visit_with_visits_no_snapshot(self):
        pass



# Copyright (C) 2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from contextlib import contextmanager
from typing import Dict, Iterable, Iterator, List, Optional, Union
import warnings

import psycopg2.pool

from swh.model.model import Origin, OriginVisit, OriginVisitStatus
from swh.storage import get_storage
from swh.storage.exc import BlockedOriginException
from swh.storage.interface import StorageInterface
from swh.storage.metrics import DifferentialTimer
from swh.storage.proxies.blocking.db import BlockingState

from .db import BlockingQuery

BLOCKING_OVERHEAD_METRIC = "swh_storage_blocking_overhead_seconds"


def get_datastore(cls, db=None, blocking_db=None, **kwargs):
    assert cls in ("postgresql", "blocking")
    from .db import BlockingAdmin

    if db is None:
        db = blocking_db
    return BlockingAdmin.connect(db)


def blocking_overhead_timer(method_name: str) -> DifferentialTimer:
    """Return a properly setup DifferentialTimer for ``method_name`` of the storage"""
    return DifferentialTimer(BLOCKING_OVERHEAD_METRIC, tags={"endpoint": method_name})


class BlockingProxyStorage:
    """Blocking storage proxy

    This proxy prevents visits from a known list of origins to be performed at all.

    It uses a specific PostgreSQL database (which for now is colocated with the
    swh.storage PostgreSQL database), the access to which is implemented in the
    :mod:`.db` submodule.

    Sample configuration

    .. code-block: yaml

        storage:
          cls: blocking
          db: 'dbname=swh-blocking-proxy'
          max_pool_conns: 10
          storage:
          - cls: remote
            url: http://storage.internal.staging.swh.network:5002/

    """

    def __init__(
        self,
        storage: Union[Dict, StorageInterface],
        db: Optional[str] = None,
        blocking_db: Optional[str] = None,
        min_pool_conns: int = 1,
        max_pool_conns: int = 5,
    ):
        if db is None:
            assert blocking_db is not None
            warnings.warn(
                "'blocking_db' field in the blocking storage configuration "
                "was renamed 'db' field",
                DeprecationWarning,
            )
            db = blocking_db
        self.storage: StorageInterface = (
            get_storage(**storage) if isinstance(storage, dict) else storage
        )

        self._blocking_pool = psycopg2.pool.ThreadedConnectionPool(
            min_pool_conns, max_pool_conns, db
        )

    def origin_visit_status_add(
        self, visit_statuses: List[OriginVisitStatus]
    ) -> Dict[str, int]:
        with self._blocking_query() as q:
            statuses = q.origins_are_blocked([v.origin for v in visit_statuses])
            if statuses and any(
                status.state != BlockingState.NON_BLOCKED
                for status in statuses.values()
            ):
                raise BlockedOriginException(statuses)
        return self.storage.origin_visit_status_add(visit_statuses)

    def origin_visit_add(self, visits: List[OriginVisit]) -> Iterable[OriginVisit]:
        with self._blocking_query() as q:
            statuses = q.origins_are_blocked([v.origin for v in visits])
            if statuses and any(
                status.state != BlockingState.NON_BLOCKED
                for status in statuses.values()
            ):
                raise BlockedOriginException(statuses)
        return self.storage.origin_visit_add(visits)

    def origin_add(self, origins: List[Origin]) -> Dict[str, int]:
        with self._blocking_query() as q:
            statuses = {}
            for origin in origins:
                status = q.origin_is_blocked(origin.url)
                if status and status.state != BlockingState.NON_BLOCKED:
                    statuses[origin.url] = status
            if statuses:
                raise BlockedOriginException(statuses)
        return self.storage.origin_add(origins)

    @contextmanager
    def _blocking_query(self) -> Iterator[BlockingQuery]:
        ret = None
        try:
            ret = BlockingQuery.from_pool(self._blocking_pool)
            yield ret
        finally:
            if ret:
                ret.put_conn()

    def __getattr__(self, key):
        method = getattr(self.storage, key)
        if method:
            # Avoid going through __getattr__ again next time
            setattr(self, key, method)
            return method

        # Raise a NotImplementedError to make sure we don't forget to add
        # masking to any new storage functions
        raise NotImplementedError(key)



# Copyright (C) 2019-2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import attr
import pytest

from swh.model.hashutil import hash_to_hex
from swh.storage import get_storage
from swh.storage.exc import StorageArgumentException


@pytest.fixture
def swh_storage():
    storage_config = {
        "cls": "pipeline",
        "steps": [
            {"cls": "validate"},
            {"cls": "memory"},
        ],
    }

    return get_storage(**storage_config)


def test_validating_proxy_storage_content(swh_storage, sample_data):
    sample_content = sample_data.content

    content = swh_storage.content_get_data(sample_content.sha1)
    assert content is None

    with pytest.raises(StorageArgumentException, match="hashes"):
        s = swh_storage.content_add([attr.evolve(sample_content, sha1=b"a" * 20)])

    content = swh_storage.content_get_data(sample_content.sha1)
    assert content is None

    s = swh_storage.content_add([sample_content])
    assert s == {
        "content:add": 1,
        "content:add:bytes": sample_content.length,
    }

    content = swh_storage.content_get_data(sample_content.sha1)
    assert content is not None


def test_validating_proxy_storage_skipped_content(swh_storage, sample_data):
    sample_content = sample_data.skipped_content
    sample_content = attr.evolve(sample_content, sha1=b"a" * 20)
    sample_content_dict = sample_content.to_dict()

    s = swh_storage.skipped_content_add([sample_content])

    content = list(swh_storage.skipped_content_missing([sample_content_dict]))
    assert content == []

    s = swh_storage.skipped_content_add([sample_content])
    assert s == {
        "skipped_content:add": 0,
    }


def test_validating_proxy_storage_directory(swh_storage, sample_data):
    sample_directory = sample_data.directory
    id_ = hash_to_hex(sample_directory.id)

    assert swh_storage.directory_missing([sample_directory.id]) == [sample_directory.id]

    with pytest.raises(StorageArgumentException, match=f"should be {id_}"):
        s = swh_storage.directory_add([attr.evolve(sample_directory, id=b"a" * 20)])

    assert swh_storage.directory_missing([sample_directory.id]) == [sample_directory.id]

    s = swh_storage.directory_add([sample_directory])
    assert s == {
        "directory:add": 1,
    }

    assert swh_storage.directory_missing([sample_directory.id]) == []


def test_validating_proxy_storage_revision(swh_storage, sample_data):
    sample_revision = sample_data.revision
    id_ = hash_to_hex(sample_revision.id)

    assert swh_storage.revision_missing([sample_revision.id]) == [sample_revision.id]

    with pytest.raises(StorageArgumentException, match=f"should be {id_}"):
        s = swh_storage.revision_add([attr.evolve(sample_revision, id=b"a" * 20)])

    assert swh_storage.revision_missing([sample_revision.id]) == [sample_revision.id]

    s = swh_storage.revision_add([sample_revision])
    assert s == {
        "revision:add": 1,
    }

    assert swh_storage.revision_missing([sample_revision.id]) == []


def test_validating_proxy_storage_release(swh_storage, sample_data):
    sample_release = sample_data.release
    id_ = hash_to_hex(sample_release.id)

    assert swh_storage.release_missing([sample_release.id]) == [sample_release.id]

    with pytest.raises(StorageArgumentException, match=f"should be {id_}"):
        s = swh_storage.release_add([attr.evolve(sample_release, id=b"a" * 20)])

    assert swh_storage.release_missing([sample_release.id]) == [sample_release.id]

    s = swh_storage.release_add([sample_release])
    assert s == {
        "release:add": 1,
    }

    assert swh_storage.release_missing([sample_release.id]) == []


def test_validating_proxy_storage_snapshot(swh_storage, sample_data):
    sample_snapshot = sample_data.snapshot
    id_ = hash_to_hex(sample_snapshot.id)

    assert swh_storage.snapshot_missing([sample_snapshot.id]) == [sample_snapshot.id]

    with pytest.raises(StorageArgumentException, match=f"should be {id_}"):
        s = swh_storage.snapshot_add([attr.evolve(sample_snapshot, id=b"a" * 20)])

    assert swh_storage.snapshot_missing([sample_snapshot.id]) == [sample_snapshot.id]

    s = swh_storage.snapshot_add([sample_snapshot])
    assert s == {
        "snapshot:add": 1,
    }

    assert swh_storage.snapshot_missing([sample_snapshot.id]) == []



# Copyright (C) 2019-2021 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from unittest.mock import patch

import pytest

from swh.core.pytest_plugin import RPCTestAdapter
from swh.storage import get_storage
from swh.storage.api import client, server
from swh.storage.in_memory import InMemoryStorage
from swh.storage.postgresql.storage import Storage as DbStorage
from swh.storage.proxies.buffer import BufferingProxyStorage
from swh.storage.proxies.filter import FilteringProxyStorage
from swh.storage.proxies.retry import RetryingProxyStorage

STORAGES = [
    pytest.param(cls, real_class, kwargs, id=cls)
    for (cls, real_class, kwargs) in [
        ("remote", client.RemoteStorage, {"url": "url"}),
        ("memory", InMemoryStorage, {}),
        (
            "postgresql",
            DbStorage,
            {"db": "postgresql://db", "objstorage": {"cls": "memory"}},
        ),
        ("filter", FilteringProxyStorage, {"storage": {"cls": "memory"}}),
        ("buffer", BufferingProxyStorage, {"storage": {"cls": "memory"}}),
        ("retry", RetryingProxyStorage, {"storage": {"cls": "memory"}}),
    ]
]


@pytest.mark.parametrize("cls,real_class,args", STORAGES)
@patch("swh.storage.postgresql.storage.psycopg2.pool")
def test_get_storage(mock_pool, cls, real_class, args):
    """Instantiating an existing storage should be ok"""
    mock_pool.ThreadedConnectionPool.return_value = None
    actual_storage = get_storage(cls, **args)
    assert actual_storage is not None
    assert isinstance(actual_storage, real_class)


@pytest.mark.parametrize("cls,real_class,args", STORAGES)
@patch("swh.storage.postgresql.storage.psycopg2.pool")
def test_get_storage_legacy_args(mock_pool, cls, real_class, args):
    """Instantiating an existing storage should be ok even with the legacy
    explicit 'args' keys

    """
    mock_pool.ThreadedConnectionPool.return_value = None
    with pytest.warns(DeprecationWarning):
        actual_storage = get_storage(cls, args=args)
        assert actual_storage is not None
        assert isinstance(actual_storage, real_class)


def test_get_storage_failure():
    """Instantiating an unknown storage should raise"""
    with pytest.raises(ValueError, match="Unknown storage class `unknown`"):
        get_storage("unknown")


def test_get_storage_pipeline():
    config = {
        "cls": "pipeline",
        "steps": [
            {
                "cls": "filter",
            },
            {
                "cls": "buffer",
                "min_batch_size": {
                    "content": 10,
                },
            },
            {
                "cls": "memory",
            },
        ],
    }

    storage = get_storage(**config)

    assert isinstance(storage, FilteringProxyStorage)
    assert isinstance(storage.storage, BufferingProxyStorage)
    assert isinstance(storage.storage.storage, InMemoryStorage)


def test_get_storage_pipeline_legacy_args():
    config = {
        "cls": "pipeline",
        "steps": [
            {
                "cls": "filter",
            },
            {
                "cls": "buffer",
                "args": {
                    "min_batch_size": {
                        "content": 10,
                    },
                },
            },
            {
                "cls": "memory",
            },
        ],
    }

    with pytest.warns(DeprecationWarning):
        storage = get_storage(**config)

    assert isinstance(storage, FilteringProxyStorage)
    assert isinstance(storage.storage, BufferingProxyStorage)
    assert isinstance(storage.storage.storage, InMemoryStorage)


# get_storage's check_config argument tests


# the "remote" and "pipeline" cases are tested in dedicated test functions below
@pytest.mark.parametrize(
    "cls,real_class,kwargs",
    [x for x in STORAGES if x.id not in ("remote", "local", "postgresql")],
)
def test_get_storage_check_config(cls, real_class, kwargs, monkeypatch):
    """Instantiating an existing storage with check_config should be ok"""
    check_backend_check_config(monkeypatch, dict(cls=cls, **kwargs))


@patch("swh.storage.postgresql.storage.psycopg2.pool")
@pytest.mark.parametrize("clazz", ["local", "postgresql"])
def test_get_storage_local_check_config(mock_pool, monkeypatch, clazz):
    """Instantiating a local storage with check_config should be ok"""
    mock_pool.ThreadedConnectionPool.return_value = None
    check_backend_check_config(
        monkeypatch,
        {"cls": clazz, "db": "postgresql://db", "objstorage": {"cls": "memory"}},
        backend_storage_cls=DbStorage,
    )


def test_get_storage_pipeline_check_config(monkeypatch):
    """Test that the check_config option works as intended for a pipelined storage"""
    config = {
        "cls": "pipeline",
        "steps": [
            {
                "cls": "filter",
            },
            {
                "cls": "buffer",
                "min_batch_size": {
                    "content": 10,
                },
            },
            {
                "cls": "memory",
            },
        ],
    }
    check_backend_check_config(
        monkeypatch,
        config,
    )


def test_get_storage_remote_check_config(monkeypatch):
    """Test that the check_config option works as intended for a remote storage"""

    monkeypatch.setattr(
        server, "storage", get_storage(cls="memory", journal_writer={"cls": "memory"})
    )
    test_client = server.app.test_client()

    class MockedRemoteStorage(client.RemoteStorage):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.session.adapters.clear()
            self.session.mount("mock://", RPCTestAdapter(test_client))

    monkeypatch.setattr(client, "RemoteStorage", MockedRemoteStorage)

    config = {
        "cls": "remote",
        "url": "mock://example.com",
    }
    check_backend_check_config(
        monkeypatch,
        config,
    )


def check_backend_check_config(
    monkeypatch, config, backend_storage_cls=InMemoryStorage
):
    """Check the staged/indirect storage (pipeline or remote) works
    as desired with regard to the check_config option of the get_storage()
    factory function.

    If set, the check_config argument is used to call the Storage.check_config() at
    instantiation time in the get_storage() factory function. This is supposed to be
    passed through each step of the Storage pipeline until it reached the actual
    backend's (typically in memory or local) check_config() method which will perform
    the verification for read/write access to the backend storage.

    monkeypatch is supposed to be the monkeypatch pytest fixture to be used from the
    calling test_ function.

    config is the config dict passed to get_storage()

    backend_storage_cls is the class of the backend storage to be mocked to
    simulate the check_config behavior; it should then be the class of the
    actual backend storage defined in the `config`.
    """
    access = None

    def mockcheck(self, check_write=False):
        if access == "none":
            return False
        if access == "read":
            return check_write is False
        if access == "write":
            return True

    monkeypatch.setattr(backend_storage_cls, "check_config", mockcheck)

    # simulate no read nor write access to the underlying (memory) storage
    access = "none"
    # by default, no check, so no complain
    assert get_storage(**config)
    # if asked to check, complain
    with pytest.raises(EnvironmentError):
        get_storage(check_config={"check_write": False}, **config)
    with pytest.raises(EnvironmentError):
        get_storage(check_config={"check_write": True}, **config)

    # simulate no write access to the underlying (memory) storage
    access = "read"
    # by default, no check so no complain
    assert get_storage(**config)
    # if asked to check for read access, no complain
    get_storage(check_config={"check_write": False}, **config)
    # if asked to check for write access, complain
    with pytest.raises(EnvironmentError):
        get_storage(check_config={"check_write": True}, **config)

    # simulate read & write access to the underlying (memory) storage
    access = "write"
    # by default, no check so no complain
    assert get_storage(**config)
    # if asked to check for read access, no complain
    get_storage(check_config={"check_write": False}, **config)
    # if asked to check for write access, no complain
    get_storage(check_config={"check_write": True}, **config)



# Copyright (C) 2018-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# flake8: noqa

import unittest
from unittest.mock import patch

import pytest

from swh.model.hashutil import hash_to_bytes
from swh.storage.algos import diff

from .test_dir_iterator import DirectoryModel


def test__get_rev(swh_storage, sample_data):
    revision = sample_data.revision

    # does not exist then raises
    with pytest.raises(AssertionError):
        diff._get_rev(swh_storage, revision.id)

    # otherwise, we retrieve its dict representation
    swh_storage.revision_add([revision])
    actual_revision = diff._get_rev(swh_storage, revision.id)
    assert actual_revision == revision.to_dict()


@patch("swh.storage.algos.diff._get_rev")
@patch("swh.storage.algos.dir_iterators._get_dir")
class TestDiffRevisions(unittest.TestCase):
    def diff_revisions(
        self,
        rev_from,
        rev_to,
        from_dir_model,
        to_dir_model,
        expected_changes,
        mock_get_dir,
        mock_get_rev,
    ):
        rev_from_bytes = hash_to_bytes(rev_from)
        rev_to_bytes = hash_to_bytes(rev_to)

        def _get_rev(*args, **kwargs):
            if args[1] == rev_from_bytes:
                return {"directory": from_dir_model["target"]}
            else:
                return {"directory": to_dir_model["target"]}

        def _get_dir(*args, **kwargs):
            from_dir = from_dir_model.get_hash_data(args[1])
            to_dir = to_dir_model.get_hash_data(args[1])
            return from_dir if from_dir != None else to_dir

        mock_get_rev.side_effect = _get_rev
        mock_get_dir.side_effect = _get_dir

        changes = diff.diff_revisions(
            None, rev_from_bytes, rev_to_bytes, track_renaming=True
        )

        self.assertEqual(changes, expected_changes)

    def test_insert_delete(self, mock_get_dir, mock_get_rev):
        rev_from = "898ff03e1e7925ecde3da66327d3cdc7e07625ba"
        rev_to = "647c3d381e67490e82cdbbe6c96e46d5e1628ce2"

        from_dir_model = DirectoryModel()

        to_dir_model = DirectoryModel()
        to_dir_model.add_file(b"file1", "ea15f54ca215e7920c60f564315ebb7f911a5204")
        to_dir_model.add_file(b"file2", "3e5faecb3836ffcadf82cc160787e35d4e2bec6a")
        to_dir_model.add_file(b"file3", "2ae33b2984974d35eababe4890d37fbf4bce6b2c")

        expected_changes = [
            {
                "type": "insert",
                "from": None,
                "from_path": None,
                "to": to_dir_model.get_path_data(b"file1"),
                "to_path": b"file1",
            },
            {
                "type": "insert",
                "from": None,
                "from_path": None,
                "to": to_dir_model.get_path_data(b"file2"),
                "to_path": b"file2",
            },
            {
                "type": "insert",
                "from": None,
                "from_path": None,
                "to": to_dir_model.get_path_data(b"file3"),
                "to_path": b"file3",
            },
        ]

        self.diff_revisions(
            rev_from,
            rev_to,
            from_dir_model,
            to_dir_model,
            expected_changes,
            mock_get_dir,
            mock_get_rev,
        )

        from_dir_model = DirectoryModel()
        from_dir_model.add_file(b"file1", "ea15f54ca215e7920c60f564315ebb7f911a5204")
        from_dir_model.add_file(b"file2", "3e5faecb3836ffcadf82cc160787e35d4e2bec6a")
        from_dir_model.add_file(b"file3", "2ae33b2984974d35eababe4890d37fbf4bce6b2c")

        to_dir_model = DirectoryModel()

        expected_changes = [
            {
                "type": "delete",
                "from": from_dir_model.get_path_data(b"file1"),
                "from_path": b"file1",
                "to": None,
                "to_path": None,
            },
            {
                "type": "delete",
                "from": from_dir_model.get_path_data(b"file2"),
                "from_path": b"file2",
                "to": None,
                "to_path": None,
            },
            {
                "type": "delete",
                "from": from_dir_model.get_path_data(b"file3"),
                "from_path": b"file3",
                "to": None,
                "to_path": None,
            },
        ]

        self.diff_revisions(
            rev_from,
            rev_to,
            from_dir_model,
            to_dir_model,
            expected_changes,
            mock_get_dir,
            mock_get_rev,
        )

    def test_onelevel_diff(self, mock_get_dir, mock_get_rev):
        rev_from = "898ff03e1e7925ecde3da66327d3cdc7e07625ba"
        rev_to = "647c3d381e67490e82cdbbe6c96e46d5e1628ce2"

        from_dir_model = DirectoryModel()
        from_dir_model.add_file(b"file1", "ea15f54ca215e7920c60f564315ebb7f911a5204")
        from_dir_model.add_file(b"file2", "f4a96b2000be83b61254d107046fa9777b17eb34")
        from_dir_model.add_file(b"file3", "d3c00f9396c6d0277727cec522ff6ad1ea0bc2da")

        to_dir_model = DirectoryModel()
        to_dir_model.add_file(b"file2", "3ee0f38ee0ea23cc2c8c0b9d66b27be4596b002b")
        to_dir_model.add_file(b"file3", "d3c00f9396c6d0277727cec522ff6ad1ea0bc2da")
        to_dir_model.add_file(b"file4", "40460b9653b1dc507e1b6eb333bd4500634bdffc")

        expected_changes = [
            {
                "type": "delete",
                "from": from_dir_model.get_path_data(b"file1"),
                "from_path": b"file1",
                "to": None,
                "to_path": None,
            },
            {
                "type": "modify",
                "from": from_dir_model.get_path_data(b"file2"),
                "from_path": b"file2",
                "to": to_dir_model.get_path_data(b"file2"),
                "to_path": b"file2",
            },
            {
                "type": "insert",
                "from": None,
                "from_path": None,
                "to": to_dir_model.get_path_data(b"file4"),
                "to_path": b"file4",
            },
        ]

        self.diff_revisions(
            rev_from,
            rev_to,
            from_dir_model,
            to_dir_model,
            expected_changes,
            mock_get_dir,
            mock_get_rev,
        )

    def test_twolevels_diff(self, mock_get_dir, mock_get_rev):
        rev_from = "898ff03e1e7925ecde3da66327d3cdc7e07625ba"
        rev_to = "647c3d381e67490e82cdbbe6c96e46d5e1628ce2"

        from_dir_model = DirectoryModel()
        from_dir_model.add_file(b"file1", "ea15f54ca215e7920c60f564315ebb7f911a5204")
        from_dir_model.add_file(
            b"dir1/file1", "8335fca266811bac7ae5c8e1621476b4cf4156b6"
        )
        from_dir_model.add_file(
            b"dir1/file2", "a6127d909e79f1fcb28bbf220faf86e7be7831e5"
        )
        from_dir_model.add_file(
            b"dir1/file3", "18049b8d067ce1194a7e1cce26cfa3ae4242a43d"
        )
        from_dir_model.add_file(b"file2", "d3c00f9396c6d0277727cec522ff6ad1ea0bc2da")

        to_dir_model = DirectoryModel()
        to_dir_model.add_file(b"file1", "3ee0f38ee0ea23cc2c8c0b9d66b27be4596b002b")
        to_dir_model.add_file(b"dir1/file2", "de3548b32a8669801daa02143a66dae21fe852fd")
        to_dir_model.add_file(b"dir1/file3", "18049b8d067ce1194a7e1cce26cfa3ae4242a43d")
        to_dir_model.add_file(b"dir1/file4", "f5c3f42aec5fe7b92276196c350cbadaf4c51f87")
        to_dir_model.add_file(b"file2", "d3c00f9396c6d0277727cec522ff6ad1ea0bc2da")

        expected_changes = [
            {
                "type": "delete",
                "from": from_dir_model.get_path_data(b"dir1/file1"),
                "from_path": b"dir1/file1",
                "to": None,
                "to_path": None,
            },
            {
                "type": "modify",
                "from": from_dir_model.get_path_data(b"dir1/file2"),
                "from_path": b"dir1/file2",
                "to": to_dir_model.get_path_data(b"dir1/file2"),
                "to_path": b"dir1/file2",
            },
            {
                "type": "insert",
                "from": None,
                "from_path": None,
                "to": to_dir_model.get_path_data(b"dir1/file4"),
                "to_path": b"dir1/file4",
            },
            {
                "type": "modify",
                "from": from_dir_model.get_path_data(b"file1"),
                "from_path": b"file1",
                "to": to_dir_model.get_path_data(b"file1"),
                "to_path": b"file1",
            },
        ]

        self.diff_revisions(
            rev_from,
            rev_to,
            from_dir_model,
            to_dir_model,
            expected_changes,
            mock_get_dir,
            mock_get_rev,
        )

    def test_insert_delete_empty_dirs(self, mock_get_dir, mock_get_rev):
        rev_from = "898ff03e1e7925ecde3da66327d3cdc7e07625ba"
        rev_to = "647c3d381e67490e82cdbbe6c96e46d5e1628ce2"

        from_dir_model = DirectoryModel()
        from_dir_model.add_file(
            b"dir3/file1", "ea15f54ca215e7920c60f564315ebb7f911a5204"
        )

        to_dir_model = DirectoryModel()
        to_dir_model.add_file(b"dir3/file1", "ea15f54ca215e7920c60f564315ebb7f911a5204")
        to_dir_model.add_file(b"dir3/dir1/")

        expected_changes = [
            {
                "type": "insert",
                "from": None,
                "from_path": None,
                "to": to_dir_model.get_path_data(b"dir3/dir1"),
                "to_path": b"dir3/dir1",
            }
        ]

        self.diff_revisions(
            rev_from,
            rev_to,
            from_dir_model,
            to_dir_model,
            expected_changes,
            mock_get_dir,
            mock_get_rev,
        )

        from_dir_model = DirectoryModel()
        from_dir_model.add_file(b"dir1/dir2/")
        from_dir_model.add_file(
            b"dir1/file1", "ea15f54ca215e7920c60f564315ebb7f911a5204"
        )

        to_dir_model = DirectoryModel()
        to_dir_model.add_file(b"dir1/file1", "ea15f54ca215e7920c60f564315ebb7f911a5204")

        expected_changes = [
            {
                "type": "delete",
                "from": from_dir_model.get_path_data(b"dir1/dir2"),
                "from_path": b"dir1/dir2",
                "to": None,
                "to_path": None,
            }
        ]

        self.diff_revisions(
            rev_from,
            rev_to,
            from_dir_model,
            to_dir_model,
            expected_changes,
            mock_get_dir,
            mock_get_rev,
        )

    def test_track_renaming(self, mock_get_dir, mock_get_rev):
        rev_from = "898ff03e1e7925ecde3da66327d3cdc7e07625ba"
        rev_to = "647c3d381e67490e82cdbbe6c96e46d5e1628ce2"

        from_dir_model = DirectoryModel()
        from_dir_model.add_file(
            b"file1_oldname", "ea15f54ca215e7920c60f564315ebb7f911a5204"
        )
        from_dir_model.add_file(
            b"dir1/file1_oldname", "ea15f54ca215e7920c60f564315ebb7f911a5204"
        )
        from_dir_model.add_file(
            b"file2_oldname", "d3c00f9396c6d0277727cec522ff6ad1ea0bc2da"
        )

        to_dir_model = DirectoryModel()
        to_dir_model.add_file(
            b"dir1/file1_newname", "ea15f54ca215e7920c60f564315ebb7f911a5204"
        )
        to_dir_model.add_file(
            b"dir2/file1_newname", "ea15f54ca215e7920c60f564315ebb7f911a5204"
        )
        to_dir_model.add_file(
            b"file2_newname", "d3c00f9396c6d0277727cec522ff6ad1ea0bc2da"
        )

        expected_changes = [
            {
                "type": "rename",
                "from": from_dir_model.get_path_data(b"dir1/file1_oldname"),
                "from_path": b"dir1/file1_oldname",
                "to": to_dir_model.get_path_data(b"dir1/file1_newname"),
                "to_path": b"dir1/file1_newname",
            },
            {
                "type": "rename",
                "from": from_dir_model.get_path_data(b"file1_oldname"),
                "from_path": b"file1_oldname",
                "to": to_dir_model.get_path_data(b"dir2/file1_newname"),
                "to_path": b"dir2/file1_newname",
            },
            {
                "type": "rename",
                "from": from_dir_model.get_path_data(b"file2_oldname"),
                "from_path": b"file2_oldname",
                "to": to_dir_model.get_path_data(b"file2_newname"),
                "to_path": b"file2_newname",
            },
        ]

        self.diff_revisions(
            rev_from,
            rev_to,
            from_dir_model,
            to_dir_model,
            expected_changes,
            mock_get_dir,
            mock_get_rev,
        )



# Copyright (C) 2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from functools import partial

import pytest
from pytest_postgresql import factories

from swh.core.db.db_utils import initialize_database_for_module
from swh.storage.proxies.masking import MaskingProxyStorage
from swh.storage.proxies.masking.db import MaskingAdmin, MaskingQuery

masking_db_postgresql_proc = factories.postgresql_proc(
    load=[
        partial(
            initialize_database_for_module,
            modname="storage.proxies.masking",
            version=MaskingAdmin.current_version,
        ),
    ],
)


masking_db_postgresql = factories.postgresql(
    "masking_db_postgresql_proc",
)


@pytest.fixture
def masking_admin(masking_db_postgresql) -> MaskingAdmin:
    return MaskingAdmin.connect(masking_db_postgresql.info.dsn)


@pytest.fixture
def masking_query(masking_db_postgresql) -> MaskingQuery:
    return MaskingQuery.connect(masking_db_postgresql.info.dsn)


@pytest.fixture
def swh_storage_backend_config():
    yield {
        "cls": "memory",
        "journal_writer": {
            "cls": "memory",
        },
    }


@pytest.fixture
def swh_storage(masking_db_postgresql, swh_storage_backend):
    return MaskingProxyStorage(
        db=masking_db_postgresql.info.dsn, storage=swh_storage_backend
    )



# Copyright (C) 2015-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# WARNING: do not import unnecessary things here to keep cli startup time under
# control
import logging
import os
from typing import IO, Callable, Dict, Optional, Tuple, Union, cast

import click

from swh.core.cli import CONTEXT_SETTINGS
from swh.core.cli import swh as swh_cli_group

try:
    from systemd.daemon import notify
except ImportError:
    notify = None


@swh_cli_group.group(name="storage", context_settings=CONTEXT_SETTINGS)
@click.option(
    "--config-file",
    "-C",
    default=None,
    type=click.Path(
        exists=True,
        dir_okay=False,
    ),
    help="Configuration file.",
)
@click.option(
    "--check-config",
    default=None,
    type=click.Choice(["no", "read", "write"]),
    help=(
        "Check the configuration of the storage at startup for read or write access; "
        "if set, override the value present in the configuration file if any. "
        "Defaults to 'read' for the 'backfill' command, and 'write' for 'rpc-server' "
        "and 'replay' commands."
    ),
)
@click.pass_context
def storage(ctx, config_file, check_config):
    """Software Heritage Storage tools."""
    from swh.core import config

    if not config_file:
        config_file = os.environ.get("SWH_CONFIG_FILENAME")

    if config_file:
        if not os.path.exists(config_file):
            raise ValueError("%s does not exist" % config_file)
        conf = config.read(config_file)
    else:
        conf = {}

    if "storage" not in conf:
        ctx.fail("You must have a storage configured in your config file.")

    ctx.ensure_object(dict)
    ctx.obj["config"] = conf
    ctx.obj["check_config"] = check_config


@storage.command(name="create-keyspace")
@click.pass_context
def create_keyspace(ctx):
    """Creates a Cassandra keyspace with table definitions suitable for use
    by swh-storage's Cassandra backend"""
    from swh.storage.cassandra import create_keyspace

    config = ctx.obj["config"]["storage"]

    for key in ("cls", "hosts", "keyspace", "auth_provider"):
        if key not in config:
            ctx.fail(f"Missing {key} key in config file.")

    if config["cls"] != "cassandra":
        ctx.fail(f"cls must be 'cassandra', not '{config['cls']}'")

    create_keyspace(
        hosts=config["hosts"],
        port=config.get("port", 9042),
        keyspace=config["keyspace"],
        auth_provider=config["auth_provider"],
    )

    print("Done.")


@storage.command(name="rpc-serve")
@click.option(
    "--host",
    default="0.0.0.0",
    metavar="IP",
    show_default=True,
    help="Host ip address to bind the server on",
)
@click.option(
    "--port",
    default=5002,
    type=click.INT,
    metavar="PORT",
    show_default=True,
    help="Binding port of the server",
)
@click.option(
    "--debug/--no-debug",
    default=True,
    help="Indicates if the server should run in debug mode",
)
@click.pass_context
def serve(ctx, host, port, debug):
    """Software Heritage Storage RPC server.

    Do NOT use this in a production environment.
    """
    from swh.storage.api.server import app

    if "log_level" in ctx.obj:
        logging.getLogger("werkzeug").setLevel(ctx.obj["log_level"])
    ensure_check_config(ctx.obj["config"], ctx.obj["check_config"], "write")
    app.config.update(ctx.obj["config"])
    app.run(host, port=int(port), debug=bool(debug))


@storage.command(name="backfill")
@click.argument("object_type")
@click.option("--start-object", default=None)
@click.option("--end-object", default=None)
@click.option("--dry-run", is_flag=True, default=False)
@click.pass_context
def backfill(ctx, object_type, start_object, end_object, dry_run):
    """Run the backfiller

    The backfiller list objects from a Storage and produce journal entries from
    there.

    Typically used to rebuild a journal or compensate for missing objects in a
    journal (eg. due to a downtime of this later).

    The configuration file requires the following entries:

    - brokers: a list of kafka endpoints (the journal) in which entries will be
      added.
    - storage_dbconn: URL to connect to the storage DB.
    - prefix: the prefix of the topics (topics will be <prefix>.<object_type>).
    - client_id: the kafka client ID.

    """
    ensure_check_config(ctx.obj["config"], ctx.obj["check_config"], "read")

    # for "lazy" loading
    from swh.storage.backfill import JournalBackfiller

    try:
        from systemd.daemon import notify
    except ImportError:
        notify = None

    conf = ctx.obj["config"]
    backfiller = JournalBackfiller(conf)

    if notify:
        notify("READY=1")

    try:
        backfiller.run(
            object_type=object_type,
            start_object=start_object,
            end_object=end_object,
            dry_run=dry_run,
        )
    except KeyboardInterrupt:
        if notify:
            notify("STOPPING=1")
        ctx.exit(0)


@storage.command(name="replay")
@click.option(
    "--stop-after-objects",
    "-n",
    default=None,
    type=int,
    help="Stop after processing this many objects. Default is to " "run forever.",
)
@click.option(
    "--type",
    "-t",
    "object_types",
    default=[],
    type=click.Choice(
        # use a hardcoded list to prevent having to load the
        # replay module at cli loading time
        [
            "origin",
            "origin_visit",
            "origin_visit_status",
            "snapshot",
            "revision",
            "release",
            "directory",
            "content",
            "skipped_content",
            "metadata_authority",
            "metadata_fetcher",
            "raw_extrinsic_metadata",
            "extid",
        ]
    ),
    help="Object types to replay",
    multiple=True,
)
@click.option(
    "--known-mismatched-hashes",
    "-X",
    "invalid_hashes_file",
    default=None,
    type=click.File("r"),
    help=(
        "File of SWHIDs of objects that are known to have invalid hashes "
        "but still need to be replayed."
    ),
)
@click.pass_context
def replay(
    ctx: click.Context,
    stop_after_objects: int,
    object_types: str,
    invalid_hashes_file: IO,
):
    """Fill a Storage by reading a Journal.

    This is typically used for a mirror configuration, reading the Software
    Heritage kafka journal to retrieve objects of the Software Heritage main
    storage to feed a replication storage. There can be several 'replayers'
    filling a Storage as long as they use the same `group-id`.

    The expected configuration file should have 2 sections:

    - storage: the configuration of the storage in which to add objects
      received from the kafka journal,

    - journal_client: the configuration of access to the kafka journal. See the
      documentation of `swh.journal` for more details on the possible
      configuration entries in this section.

      https://docs.softwareheritage.org/devel/apidoc/swh.journal.client.html

    In addition to these 2 mandatory config sections, a third 'replayer' may be
    specified with a 'error_reporter' config entry allowing to specify redis
    connection parameters that will be used to report non-recoverable mirroring,
    eg.::

      storage:
        [...]
      journal_client:
        [...]
      replayer:
        error_reporter:
          host: redis.local
          port: 6379
          db: 1

    """
    import functools

    from swh.journal.client import get_journal_client
    from swh.model.swhids import CoreSWHID
    from swh.storage import get_storage
    from swh.storage.replay import ModelObjectDeserializer, process_replay_objects

    ensure_check_config(ctx.obj["config"], ctx.obj["check_config"], "write")

    conf = ctx.obj["config"]
    storage = get_storage(**conf.pop("storage"))

    client_cfg = conf.pop("journal_client")
    replayer_cfg = conf.pop("replayer", {})

    if "error_reporter" in replayer_cfg:
        from redis import Redis

        reporter = Redis(**replayer_cfg.get("error_reporter")).set
    else:
        reporter = None
    validate = client_cfg.get("privileged", False)

    if not validate and reporter:
        ctx.fail(
            "Invalid configuration: you cannot have 'error_reporter' set if "
            "'privileged' is False; we cannot validate anonymized objects."
        )

    KMH = Optional[Tuple[Tuple[str, bytes, bytes]]]

    known_mismatched_hashes: KMH = None
    if validate and invalid_hashes_file:
        inv_obj_ids = (
            row.strip().split(",") for row in invalid_hashes_file if row.strip()
        )
        swhids = (
            (CoreSWHID.from_string(swhid.strip()), bytes.fromhex(computed_id.strip()))
            for swhid, computed_id in inv_obj_ids
        )

        known_mismatched_hashes = cast(
            KMH,
            tuple(
                (swhid.object_type.name.lower(), swhid.object_id, computed_id)
                for swhid, computed_id in swhids
            ),
        )

    deserializer = ModelObjectDeserializer(
        reporter=cast(Union[Callable[[str, bytes], None], None], reporter),
        validate=validate,
        known_mismatched_hashes=known_mismatched_hashes,
    )

    client_cfg["value_deserializer"] = deserializer.convert
    if object_types:
        client_cfg["object_types"] = object_types
    if stop_after_objects:
        client_cfg["stop_after_objects"] = stop_after_objects

    try:
        client = get_journal_client(**client_cfg)
    except ValueError as exc:
        ctx.fail(str(exc))

    worker_fn = functools.partial(process_replay_objects, storage=storage)

    if notify:
        notify("READY=1")

    try:
        client.process(worker_fn)
    except KeyboardInterrupt:
        ctx.exit(0)
    else:
        print("Done.")
    finally:
        if notify:
            notify("STOPPING=1")
        client.close()


@storage.command(name="create-object-reference-partitions")
@click.argument("start")
@click.argument("end")
@click.pass_context
def create_object_reference_partitions(ctx: click.Context, start: str, end: str):
    """Create object_reference partitions from START_DATE to END_DATE"""

    import datetime

    from swh.storage import get_storage
    from swh.storage.postgresql.storage import Storage as PostgreSQLStorage

    storage = get_storage(**ctx.obj["config"]["storage"])

    # This function uses the PostgreSQL swh.storage.db attribute directly, so we
    # need unwrap all the layers of proxies (e.g. record_references, buffer,
    # filter, etc.) to get access to the "concrete" underlying storage.
    while hasattr(storage, "storage"):
        storage = storage.storage

    if not isinstance(storage, PostgreSQLStorage):
        ctx.fail("Storage instance needs to be a direct PostgreSQL storage")

    start_date = datetime.date.fromisoformat(start)
    end_date = datetime.date.fromisoformat(end)

    if start_date > end_date:
        ctx.fail("Start date must be before end date.")

    cur_date = start_date
    end_year, end_week = end_date.isocalendar()[0:2]
    while True:
        year, week = cur_date.isocalendar()[0:2]
        if (year, week) > (end_year, end_week):
            break

        with storage.db() as db:
            with db.transaction() as cur:
                monday, sunday = db.object_references_create_partition(
                    year, week, cur=cur
                )

        click.echo(
            "Created object references table for dates between %s and %s"
            % (monday, sunday)
        )

        cur_date += datetime.timedelta(days=7)


@storage.command(name="remove-old-object-reference-partitions")
@click.option(
    "--force",
    is_flag=True,
    default=False,
    help="do not ask for confirmation before removing tables",
)
@click.argument("before")
@click.pass_context
def remove_old_object_reference_partitions(
    ctx: click.Context, force: bool, before: str
):
    """Remove object_reference partitions for values older than BEFORE"""
    import datetime

    from swh.storage import get_storage
    from swh.storage.postgresql.storage import Storage as PostgreSQLStorage

    storage = get_storage(**ctx.obj["config"]["storage"])

    # This function uses the PostgreSQL swh.storage.db attribute directly, so we
    # need unwrap all the layers of proxies (e.g. record_references, buffer,
    # filter, etc.) to get access to the "concrete" underlying storage.
    while hasattr(storage, "storage"):
        storage = storage.storage

    if not isinstance(storage, PostgreSQLStorage):
        ctx.fail("Storage instance needs to be a direct PostgreSQL storage")

    before_date = datetime.datetime.fromisoformat(before)

    with storage.db() as db:
        existing_partitions = db.object_references_list_partitions()
        to_remove = []
        for partition in existing_partitions:
            # If it ends after the specified date, we should keep this partition
            # and the ones after.
            if partition.end > before_date:
                break
            to_remove.append((partition.year, partition.week))

        if len(to_remove) == 0:
            click.echo("Nothing needs to be removed.")
            ctx.exit(0)

        if len(existing_partitions) == len(to_remove):
            ctx.fail(
                "Trying to remove all existing partitions. This can’t be right, sorry."
            )

        if not force:
            click.echo(
                "We will remove the following partitions before "
                f"{before_date.strftime('%Y-%m-%d')}:"
            )
            for year, week in to_remove:
                click.echo(f"- {year:04d} week {week:02d}")
            click.confirm("Do you want to proceed?", default=False, abort=True)

        with db.transaction() as cur:
            for year, week in to_remove:
                db.object_references_drop_partition(year, week, cur)


def ensure_check_config(storage_cfg: Dict, check_config: Optional[str], default: str):
    """Helper function to inject the setting of check_config option in the storage config
    dict according to the expected default value (default value depends on the command,
    eg. backfill can be read-only).

    """
    if check_config is not None:
        if check_config == "no":
            storage_cfg.pop("check_config", None)
        else:
            storage_cfg["check_config"] = {"check_write": check_config == "write"}
    else:
        if "check_config" not in storage_cfg:
            storage_cfg["check_config"] = {"check_write": default == "write"}


def main():
    logging.basicConfig()
    return serve(auto_envvar_prefix="SWH_STORAGE")


if __name__ == "__main__":
    main()



# Copyright (C) 2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from unittest.mock import call
import uuid

import pytest

from swh.core.db.db_utils import get_database_info
from swh.storage.proxies.masking.db import (
    DuplicateRequest,
    MaskedState,
    MaskedStatus,
    MaskingAdmin,
    MaskingQuery,
    RequestNotFound,
)
from swh.storage.tests.storage_data import StorageData


def test_db_version(masking_admin: MaskingAdmin):
    dbmodule, dbversion, dbflavor = get_database_info(masking_admin.conn.dsn)
    assert dbmodule == "storage.proxies.masking"
    assert dbversion == MaskingAdmin.current_version
    assert dbflavor is None


def test_create_find_request(masking_admin: MaskingAdmin):
    created = masking_admin.create_request(slug="foo", reason="bar")

    assert created.slug == "foo"
    assert created.reason == "bar"
    assert created.id is not None
    assert created.date is not None

    assert masking_admin.find_request("foo") == created


def test_create_request_conflict(masking_admin: MaskingAdmin):
    masking_admin.create_request(slug="foo", reason="bar")

    with pytest.raises(DuplicateRequest) as exc_info:
        masking_admin.create_request(slug="foo", reason="quux")

    assert exc_info.value.args == ("foo",)


def test_find_request_not_found(masking_admin):
    slug = "notfound"
    assert masking_admin.find_request(slug=slug) is None


def test_find_request_by_id(masking_admin: MaskingAdmin):
    created = masking_admin.create_request(slug="foo", reason="bar")

    assert masking_admin.find_request_by_id(created.id) == created


NON_EXISTING_UUID: uuid.UUID = uuid.UUID("da785a27-acab-4a35-b82a-a5ae3714407c")


def test_find_request_by_id_not_found(masking_admin: MaskingAdmin):
    assert masking_admin.find_request_by_id(NON_EXISTING_UUID) is None


@pytest.fixture
def populated_masking_admin(masking_admin):
    pending_request1 = masking_admin.create_request(slug="pending1", reason="one")
    masking_admin.set_object_state(
        request_id=pending_request1.id,
        new_state=MaskedState.DECISION_PENDING,
        swhids=[StorageData.content.swhid().to_extended()],
    )
    pending_request2 = masking_admin.create_request(slug="pending2", reason="two")
    masking_admin.set_object_state(
        request_id=pending_request2.id,
        new_state=MaskedState.VISIBLE,
        swhids=[StorageData.content.swhid().to_extended()],
    )
    masking_admin.set_object_state(
        request_id=pending_request2.id,
        new_state=MaskedState.DECISION_PENDING,
        swhids=[StorageData.content2.swhid().to_extended()],
    )
    masking_admin.create_request(slug="cleared", reason="handled")
    # We add no masks to this last one
    return masking_admin


def test_get_requests_excluding_cleared_requests(populated_masking_admin):
    assert [
        (request.slug, count)
        for request, count in populated_masking_admin.get_requests(
            include_cleared_requests=False
        )
    ] == [("pending2", 2), ("pending1", 1)]


def test_get_requests_including_cleared_requests(populated_masking_admin):
    assert [
        (request.slug, count)
        for request, count in populated_masking_admin.get_requests(
            include_cleared_requests=True
        )
    ] == [("cleared", 0), ("pending2", 2), ("pending1", 1)]


def test_get_states_for_request(populated_masking_admin):
    request = populated_masking_admin.find_request("pending2")
    states = populated_masking_admin.get_states_for_request(request.id)
    assert states == {
        StorageData.content.swhid().to_extended(): MaskedState.VISIBLE,
        StorageData.content2.swhid().to_extended(): MaskedState.DECISION_PENDING,
    }


def test_get_states_for_request_not_found(masking_admin: MaskingAdmin):
    with pytest.raises(RequestNotFound) as exc_info:
        masking_admin.get_states_for_request(NON_EXISTING_UUID)
    assert exc_info.value.args == (NON_EXISTING_UUID,)


def test_find_masks(populated_masking_admin: MaskingAdmin):
    swhids = [
        StorageData.content.swhid().to_extended(),
        StorageData.content2.swhid().to_extended(),
        # This one does not exist in the masking db
        StorageData.directory.swhid().to_extended(),
    ]
    masks = populated_masking_admin.find_masks(swhids)
    # The order in the output should be grouped by SWHID
    assert [(mask.swhid, mask.state, mask.request_slug) for mask in masks] == [
        (
            StorageData.content2.swhid().to_extended(),
            MaskedState.DECISION_PENDING,
            "pending2",
        ),
        (StorageData.content.swhid().to_extended(), MaskedState.VISIBLE, "pending2"),
        (
            StorageData.content.swhid().to_extended(),
            MaskedState.DECISION_PENDING,
            "pending1",
        ),
    ]


def test_delete_masks(populated_masking_admin):
    request = populated_masking_admin.find_request("pending2")
    populated_masking_admin.delete_masks(request.id)

    assert populated_masking_admin.get_states_for_request(request.id) == {}


def test_delete_masks_not_found(masking_admin: MaskingAdmin):
    with pytest.raises(RequestNotFound) as exc_info:
        masking_admin.delete_masks(NON_EXISTING_UUID)
    assert exc_info.value.args == (NON_EXISTING_UUID,)


def test_record_history(masking_admin: MaskingAdmin):
    messages = [f"message {i}" for i in range(3)]

    request = masking_admin.create_request(slug="foo", reason="bar")

    for message in messages:
        msg = masking_admin.record_history(request.id, message)
        assert msg.date is not None


def test_record_history_not_found(masking_admin: MaskingAdmin):
    with pytest.raises(RequestNotFound) as exc_info:
        masking_admin.record_history(NON_EXISTING_UUID, "kaboom")
    assert exc_info.value.args == (NON_EXISTING_UUID,)


def test_get_history(masking_admin: MaskingAdmin):
    request = masking_admin.create_request(slug="foo", reason="bar")
    masking_admin.record_history(request.id, "one")
    masking_admin.record_history(request.id, "two")
    masking_admin.record_history(request.id, "three")

    history = masking_admin.get_history(request.id)
    assert all(record.request == request.id for record in history)
    assert ["three", "two", "one"] == [record.message for record in history]


def test_get_history_not_found(masking_admin: MaskingAdmin):
    with pytest.raises(RequestNotFound) as exc_info:
        masking_admin.get_history(NON_EXISTING_UUID)
    assert exc_info.value.args == (NON_EXISTING_UUID,)


def test_swhid_lifecycle(masking_admin: MaskingAdmin, masking_query: MaskingQuery):
    # Create a request
    request = masking_admin.create_request(slug="foo", reason="bar")

    masked_swhids = [
        StorageData.content.swhid().to_extended(),
        StorageData.directory.swhid().to_extended(),
        StorageData.revision.swhid().to_extended(),
        StorageData.release.swhid().to_extended(),
        StorageData.snapshot.swhid().to_extended(),
        StorageData.origin.swhid(),
    ]

    all_swhids = masked_swhids + [
        StorageData.content2.swhid().to_extended(),
        StorageData.directory2.swhid().to_extended(),
        StorageData.revision2.swhid().to_extended(),
        StorageData.release2.swhid().to_extended(),
        StorageData.empty_snapshot.swhid().to_extended(),
        StorageData.origin2.swhid(),
    ]

    masking_admin.set_object_state(
        request_id=request.id,
        new_state=MaskedState.DECISION_PENDING,
        swhids=masked_swhids,
    )

    expected = {
        swhid: [MaskedStatus(state=MaskedState.DECISION_PENDING, request=request.id)]
        for swhid in masked_swhids
    }

    assert masking_query.swhids_are_masked(all_swhids) == expected
    assert dict(masking_query.iter_masked_swhids()) == expected

    restricted = masked_swhids[0:2]

    masking_admin.set_object_state(
        request_id=request.id, new_state=MaskedState.RESTRICTED, swhids=restricted
    )

    for swhid in restricted:
        expected[swhid] = [
            MaskedStatus(state=MaskedState.RESTRICTED, request=request.id)
        ]

    assert masking_query.swhids_are_masked(all_swhids) == expected
    assert dict(masking_query.iter_masked_swhids()) == expected

    visible = masked_swhids[2:4]

    masking_admin.set_object_state(
        request_id=request.id, new_state=MaskedState.VISIBLE, swhids=visible
    )

    for swhid in visible:
        del expected[swhid]

    assert masking_query.swhids_are_masked(all_swhids) == expected
    assert dict(masking_query.iter_masked_swhids()) == expected


def test_set_display_name(masking_admin: MaskingAdmin, masking_query: MaskingQuery):
    assert masking_query.display_name([b"author1@example.com"]) == {}
    assert masking_query.display_name([b"author2@example.com"]) == {}

    masking_admin.set_display_name(
        b"author1@example.com", b"author2 <author2@example.com>"
    )

    assert masking_query.display_name([b"author1@example.com"]) == {
        b"author1@example.com": b"author2 <author2@example.com>"
    }
    assert masking_query.display_name([b"author2@example.com"]) == {}


def test_query_metrics(
    masking_admin: MaskingAdmin, masking_query: MaskingQuery, mocker
):
    increment = mocker.patch("swh.core.statsd.statsd.increment")

    # Create a request
    request = masking_admin.create_request(slug="foo", reason="bar")

    masked_swhids = [
        StorageData.content.swhid().to_extended(),
        StorageData.directory.swhid().to_extended(),
        StorageData.revision.swhid().to_extended(),
        StorageData.release.swhid().to_extended(),
        StorageData.snapshot.swhid().to_extended(),
        StorageData.origin.swhid(),
    ]

    all_swhids = masked_swhids + [
        StorageData.content2.swhid().to_extended(),
        StorageData.directory2.swhid().to_extended(),
        StorageData.revision2.swhid().to_extended(),
        StorageData.release2.swhid().to_extended(),
        StorageData.empty_snapshot.swhid().to_extended(),
        StorageData.origin2.swhid(),
    ]

    # Query with no masked SWHIDs

    assert masking_query.swhids_are_masked(all_swhids) == {}
    increment.assert_called_once_with(
        "swh_storage_masking_queried_total", len(all_swhids)
    )
    increment.reset_mock()

    assert dict(masking_query.iter_masked_swhids()) == {}
    increment.assert_has_calls(
        [
            call("swh_storage_masking_list_requests_total", 1),
            call("swh_storage_masking_listed_total", 0),
        ]
    )
    increment.reset_mock()

    # Mask some SWHIDs

    masking_admin.set_object_state(
        request_id=request.id,
        new_state=MaskedState.DECISION_PENDING,
        swhids=masked_swhids,
    )

    # Query again

    assert len(masking_query.swhids_are_masked(all_swhids)) == len(masked_swhids)
    increment.assert_has_calls(
        [
            call("swh_storage_masking_queried_total", len(all_swhids)),
            call("swh_storage_masking_masked_total", len(masked_swhids)),
        ]
    )
    increment.reset_mock()

    assert set(dict(masking_query.iter_masked_swhids())) == set(masked_swhids)
    increment.assert_has_calls(
        [
            call("swh_storage_masking_list_requests_total", 1),
            call("swh_storage_masking_listed_total", len(masked_swhids)),
        ]
    )



# Copyright (C) 2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


import pytest

from swh.storage.proxies.masking import MaskingProxyStorage
from swh.storage.tests.test_in_memory import TestInMemoryStorage as _TestStorage


class TestStorage(_TestStorage):
    @pytest.mark.xfail(reason="typing.Protocol instance check is annoying")
    def test_types(self, *args, **kwargs):
        super().test_types(*args, **kwargs)

    def test_storage_type(self, swh_storage):
        assert isinstance(swh_storage, MaskingProxyStorage)
